#open "prelude";;
#open "adl";;

type token_type =
  IDENT of string | INT of int
| ATOM | BAR | BEGIN | CARD | CASE | CASETYPE | COLON | COMMA | DO | DOT 
| ELSE | EMPTYSET | END | EPSILON | EQUAL 
| FORALL | FORONE | FUNCTION | GE | IF | IN
| LATOM | LBRACK | LE | LPAREN | MAX | MEASURE | MIN | NOT | OF | OTHERWISE 
| POW | PROCEDURE | RBRACK | RPAREN | SEMICOL | SLASH | THEN | TO | TYPE 
| UNDERSCORE
;;

let keywords =
  let t = hashtbl__new 13 in
  hashtbl__add t "atom" ATOM;
  hashtbl__add t "begin" BEGIN;
  hashtbl__add t "card" CARD;
  hashtbl__add t "case" CASE;
  hashtbl__add t "casetype" CASETYPE;
  hashtbl__add t "do" DO;
  hashtbl__add t "else" ELSE;
  hashtbl__add t "end" END;
  hashtbl__add t "epsilon" EPSILON;
  hashtbl__add t "forall" FORALL;
  hashtbl__add t "forone" FORONE;
  hashtbl__add t "function" FUNCTION;
  hashtbl__add t "if" IF;
  hashtbl__add t "in" IN;
  hashtbl__add t "Latom" LATOM;
  hashtbl__add t "max" MAX;
  hashtbl__add t "measure" MEASURE;
  hashtbl__add t "min" MIN;
  hashtbl__add t "not" NOT;
  hashtbl__add t "of" OF;
  hashtbl__add t "otherwise" OTHERWISE;
  hashtbl__add t "procedure" PROCEDURE;
  hashtbl__add t "then" THEN;
  hashtbl__add t "to" TO;
  hashtbl__add t "type" TYPE;
  t
;;

let token_name = function
  IDENT _ -> "IDENT" | INT _ -> "INT"
| ATOM -> "Atom" | BAR -> "|" | BEGIN -> "begin" | CARD -> "card"
| CASE -> "case" | CASETYPE -> "casetype" 
| COLON -> ":" 
| COMMA -> "," | DO -> "do" | DOT -> "." 
| ELSE -> "else" | EMPTYSET -> "{}" | END -> "end" | EPSILON -> "epsilon"
| EQUAL -> "=" | FORALL -> "forall" | FORONE -> "forone" 
| FUNCTION -> "function" | GE -> ">="
| IF -> "if" | IN -> "in" | LATOM -> "LAtom" | LBRACK -> "[" | LE -> "<="
| LPAREN -> "(" 
| MAX -> "max" | MEASURE -> "measure" | MIN -> "min" | NOT -> "not" 
| OF -> "of" | OTHERWISE -> "otherwise" | POW -> "^"
| PROCEDURE -> "procedure" | RBRACK -> "]" | RPAREN -> ")" | SEMICOL -> ";" 
| SLASH -> "/" | THEN -> "then" | TO -> "to" | TYPE -> "type"
| UNDERSCORE -> "_"
;;

let buff = create_string 1000;;

let rec ident len = function
  [< '(`a`..`z` | `A` .. `Z` | `0` .. `9` | `_` | `'`) as c; s >] ->
    set_nth_char buff len c; ident (succ len) s
| [< >] ->
    let str = sub_string buff 0 len in
    (try hashtbl__find keywords str with _ -> IDENT str)
;;

let rec number n = function
  [< '`0` .. `9` as d; s >] ->
    number(10*n+int_of_char d-int_of_char`0`) s
| [< >] -> n
;;

let ne_list p = ne_list_p
     where rec ne_list_p = function
       [< p x; list l >] -> x::l
     and list = function
       [< ne_list_p l >] -> l
     | [< >] -> []
;;

let ne_comma_list p = ne_list_p
  where rec ne_list_p = function
       [< p x; list l >] -> x::l
  and list = function
       [< 'COMMA; ne_list_p l >] -> l
     | [< >] -> []
  ;;

let rec comma_list p = function
  [< (ne_comma_list p) l >] -> l
| [< >] -> []
;;

let ne_semicol_list p = ne_list_p	(* p ; p ; ... ; p [;| ] *)
  where rec ne_list_p = function
       [< p x; list l >] -> x::l
  and list = function
       [< 'SEMICOL; list_s l >] -> l
     | [< >] -> []
  and list_s = function
       [< ne_list_p l >] -> l
     | [< >] -> []
  ;;

let rec next_token = function
  [< '(`a`..`z` | `A` .. `Z`) as c; s >] ->
    set_nth_char buff 0 c; ident 1 s
| [< '`0` .. `9` as d; s >] ->
    INT(number (int_of_char d-int_of_char `0`) s)
| [< '` ` | `\n` | `\t`; s >] -> next_token s
| [< '`%`; s >] -> skip_comment s
| [< '`|` >] -> BAR
| [< '`:` >] -> COLON
| [< '`,` >] -> COMMA
| [< '`.` >] -> DOT
| [< '`=` >] -> EQUAL
| [< '`>`; '`=` >] -> GE
| [< '`[` >] -> LBRACK
| [< '`<`; '`=` >] -> LE
| [< '`(` >] -> LPAREN
| [< '`^` >] -> POW
| [< '`]` >] -> RBRACK
| [< '`)` >] -> RPAREN
| [< '`;` >] -> SEMICOL
| [< '`/` >] -> SLASH
| [< '`{`; '`}` >] -> EMPTYSET
| [< '`_` >] -> UNDERSCORE
| [< 'x >] -> failwith ("Bad char: "^make_string 1 x)

and skip_comment = function
  [< '`%`; s >] -> next_token s
| [< 'x; s >] -> skip_comment s
;;

let rec reset_lexer = function
  [< '`\n` >] -> ()
| [< '_; reset_lexer _ >] -> ()
| [< >] -> ()
;;

let rec adl_file = function
    [< 'TYPE; (ne_list type_decl) t; (list f_decl) f; meas_decls m >]
	-> File(flat t,f,m)

and f_decl = function
    [< 'FUNCTION; function_decl f >] -> f
  | [< 'PROCEDURE; procedure_decl p >] -> p

and meas_decls = function
    [< 'MEASURE; (ne_list meas_decl) m >] -> flat m
  | [< >] -> []

and type_decl = function
    [< (ne_comma_list id) d1; 'EQUAL; struct d3; 'SEMICOL >] -> distrib d3 d1

and struct = function
    [< (ne_list struct2) d1; type_decl2 d2 >] 
	-> let newd1 = transform("product",d1) in
		if d2=Empty then newd1 else Union(newd1,d2)

and type_decl2 = function
    [< 'BAR; struct d3 >] -> d3
  | [< >] -> Empty

and id = function [< 'IDENT d1 >] -> d1

and struct2 = function
    [< 'IDENT d1; (struct3 d1) d2 >] -> d2
  | [< 'ATOM; struct_atom y >] -> y
  | [< 'LATOM; struct_latom y >] -> y
  | [< 'LBRACK; 'IDENT s; 'RBRACK >] -> add_mark s; Mark s
  | [< 'EPSILON >] -> Epsilon
  | [< 'MAX; 'LPAREN; 'IDENT d3; 'RPAREN >] -> Ident (d3,Max)
  | [< 'MIN; 'LPAREN; 'IDENT d3; 'RPAREN >] -> Ident (d3,Min)
  | [< 'LPAREN; struct_lp d2 >] -> d2

and struct_lp = function
    [< 'RPAREN >] -> Epsilon
  | [< struct d2; 'RPAREN >] -> d2

and struct_atom = function
    [< 'LPAREN; atom_field d3; 'RPAREN >] -> Atom d3
  | [< >] -> Atom (1,1,"z")

and struct_latom = function
    [< 'LPAREN; atom_field d3; 'RPAREN >] -> LAtom d3
  | [< >] -> LAtom (1,1,"z")

and struct3 d1 = function
    [< 'POW >] -> Ident (d1,Max)
  | [< 'UNDERSCORE >] -> Ident (d1,Min)
  | [< 'LPAREN; struct d3; (struct4 d1 d3) r >] -> r
  | [< >] -> Ident (d1,No)

and struct4 d1 d3 = function
    [< 'RPAREN >] -> transform (d1,[d3])
  | [< 'COMMA; (struct5 d1 d3) r >] -> r

and struct5 d1 d3 = function
    [< 'CARD; card d6; 'RPAREN >] -> transform_card (d1,d3,d6)
  | [< (ne_comma_list struct) d5; 'RPAREN >] -> transform (d1,d3::d5)
        
and atom_field = function
    [< rat d1; (atom_field2 d1) d2 >] -> d2
        
and atom_field2 d1 = function
    [< 'COMMA; (atom_field3 d1) y >] -> y
  | [< >] -> (d1,1,"z")
        
and atom_field3 d1 = function
    [< 'IDENT d3 >] -> (d1,1,d3)
  | [< rat d3; (atom_field4 d1 d3) y >] -> y
        
and atom_field4 d1 d3 = function
    [< 'COMMA; 'IDENT d5 >] -> (d1,d3,d5)
  | [< >] -> (d1,d3,"z")
        
and rat = function
    [< 'INT d1; (rat2 d1) d2 >] -> d2
        
and rat2 d1 = function
    [< 'SLASH; 'INT d3 >] -> d1/d3
  | [< >] -> d1
        
and card = function
    [< 'EQUAL; 'INT d2 >] -> Eq d2
  | [< 'GE; 'INT d2 >] -> Ge (All,d2)
  | [< 'LE; 'INT d2 >] -> Le d2
        
and meas_decl = function
    [< (ne_comma_list id) d1; 'COLON; cost d3; 'SEMICOL >] -> distrib d3 d1
        
and cost = function
    [< 'INT d1 >] -> NumCost d1
        | [< 'IDENT d1 >] -> ParCost d1
        
and function_decl = function
    [< 'IDENT d2; 'LPAREN; (ne_comma_list fpar) d4; 'RPAREN; 'COLON; 'IDENT d7;
	'SEMICOL; AdlInstr d9; 'SEMICOL >] -> (d2,(flat d4,(d7,d9)))

and procedure_decl = function
    [< 'IDENT d2; 'LPAREN; (ne_semicol_list fpar) d4; 'RPAREN; 'SEMICOL; AdlInstr d7; 'SEMICOL >] -> (d2,(flat d4,("",d7)))

and fpar = function
    [< (ne_comma_list id) d1; 'COLON; struct2 d3 >] -> distrib d3 d1
        
and AdlInstr = function
    [< 'CASE; (ne_comma_list id) d2; 'OF; (ne_semicol_list one_case) d4; 
	'END >] -> Cases (d2,d4)
  | [< 'CASETYPE; 'IDENT d2; 'OF; (ne_semicol_list one_case) d4; 'END >] 
	-> Casetype (d2,d4)
  | [< 'BEGIN; (ne_semicol_list AdlInstr) d2; 'END >] -> Instructions d2
  | [< 'IF; condition d2; 'THEN; AdlInstr d4; (if_then d2 d4) y >] -> y
  | [< 'FORALL; 'IDENT d2; 'IN; 'IDENT d4; 'DO; AdlInstr d6 >] 
	-> Forall (d2,d4,d6)
  | [< 'FORONE; 'IDENT d2; 'IN; 'IDENT d4; 'DO; AdlInstr d6 >] 
	-> Forone (d2,d4,d6)
  | [< 'TO; repeat d2; 'DO; AdlInstr d4 >] -> Repeat (d4,d2)
  | [< expression d1 >] -> Expression d1
  | [< >] -> Empty_instr

and semicol = function [< 'SEMICOL >] -> () | [< >] -> ()

and if_then d2 d4 = function
    [< 'ELSE; AdlInstr d6 >] -> IfThenElse (d2,d4,d6)
  | [< >] -> IfThenElse (d2,d4,Empty_instr)
        
and repeat = function
    [< 'INT d1 >] -> RepNum d1
        | [< 'IDENT d1 >] -> RepId d1
        
and condition = function
    [< 'IDENT d1; (condition2 d1) y >] -> y
        
and condition2 d1 = function
    [< 'EQUAL; pattern d3 >] -> Equal(Ident(d1,No),d3)
  | [< 'LPAREN; 'IDENT d3; 'RPAREN >] -> Predicate (d1,d3)
        
and expression = function
    [< 'IDENT d1; (expression2 d1) d2 >] -> d2

and expression2 d1 = function
    [< 'LPAREN; (ne_comma_list expression) d3; 'RPAREN >] 
	-> if is_constr d1 then Exp_constr (d1,d3) else Call (d1,d3)
  | [< >] -> Exp_id d1
          
and one_case = function
    [< (ne_comma_list pattern) d1; 'COLON; AdlInstr d3 >] -> (d1,d3)
        
and pattern = function
    [< 'IDENT d1; (pattern2 d1) d2 >] -> d2
  | [< 'EPSILON >] -> Pat_epsilon
  | [< 'LPAREN; (comma_list pattern) d2; 'RPAREN >] -> transform1 d2
  | [< 'OTHERWISE >] -> Pat_all
  | [< 'EMPTYSET >] -> Pat_emptyset

and pattern2 d1 = function
    [< 'LPAREN; (ne_comma_list pattern) d3; 'RPAREN >] 
	-> transform1 (Pat_id d1::d3)
  | [< >] -> Pat_id d1
;;



