#open "prelude";;
#open "output";;
#open "adl";;
#open "gf";;
#open "types";;
#open "comp";;
#open "analyze";;
#open "formulae";;
#open "predicate";;

let depth = ref 0;;

let cd_eqs = ref [("tau_name999",Z)];;
(* equations for already computed complexity descriptors and cond.g.f. *)

let tau f = "tau_" ^ f;;

let compl_descr f ts = try_find (fun (cd,(name,strs)) -> if (name=f) & 
(for_all2 (fun x y -> eq_type x y) strs ts) then Gf_id (tau cd)
 else failwith "try_find") !cdl;;

let in_forall = ref ["xyxy"];;

let subs_compl gftype env summants vars c =
 let env1 = (map (fun v -> (v,Epsilon)) vars)@env in
  gf1 [(hd vars,c)] gftype env1 (tr_prod summants);;
(* let env1 = (map (fun v -> (v,Epsilon)) vars)@env in
  Times(gf gftype env1 (tr_prod summants),c);; *)

(************************ compl_call **********************)
let rec compl_call gftype env summants f pars = 
let v = flat(map vars_of_expr pars) in
if has_double v then err "compl_call" "same variable at <> places" v
else match pars with
  []        -> err "compl_call" "empty list of parameters for function " f
| [Call(g,l)] -> let s = name_of_composition [f;g] in 
		    reduce_compositions();
		    compl_call gftype env summants s l
| [Exp_constr _]-> compl gftype env summants (compute_value_expr (map fst env)
			(Call(f,pars)))
| [p]	    -> let t1 = eval env (struct_of_expr p)
	       and (atoms,vars) = partition (fun x -> mem x (map fst !types))
					    (vars_of_expr p) in
	       let parf = params f
	       and po = (match t1 with Ident(_,p) -> p | _ -> No) in
	       let l = (hd vars,Ident("tau_"^f,po))::
			(map (fun x -> (x,Epsilon)) (tl vars)) in
	       if list_length parf<>1 
		then err "compl_call" "bad number of parameters" parf
	        else let (x,t) = hd parf in
	        if eq_type (eval env t) (eval env t1)
		 then Ratio(gf gftype (l@env) (tr_prod summants),
			    Timess(map (fun y -> Gf_id y) atoms))
			(* do not use subs_compl because f(x^) -> int(tau_f) *)
    	         else (incr depth;if !printlevel>=4 then print_newline();
			let c = try compl_descr f [t1] with _ ->
			(let gn = gen_cd_name f in
			 cdl:=(gn,(f,[t1]))::!cdl; (* we start to analyze f over t1 *)
			 mess 2 ("Introducing the new complexity descriptor tau_"^gn^" over "^(string_of_struct t1));
			 let cc = compl gftype (map2 (fun x y -> join env x y) (params f) pars)
			  (map fst (params f)) (body f) in 
				cd_eqs:=(tau gn,cc)::!cd_eqs; cc)
		  in decr depth;subs_compl gftype env summants vars c)
| l         -> let parf = params f in
		if list_length parf <> list_length l then
		 err "compl_call" "bad number of arguments" parf
		else if exists (function Exp_id _ -> false | _ -> true) l then
		 nyi "compl_call" "other than ident"
		else let l1 = map (function Exp_id x -> x
	        | _ -> err "compl_call" "all parameters must be idents" l) l in
		      let l2 = map (fun x -> eval env (Ident(x,No))) l1 
		      and vars=filter_neg (fun x->mem x (map fst !types)) l1 in
		      if for_all2 (fun x y -> eq_type (snd x) y) parf l2
			then gf gftype ((hd l1,Ident("tau_"^f,No))::
					(map (fun x -> (x,Epsilon)) (tl l1))@
					env) (tr_prod summants)
		       else let c = try compl_descr f l2 with _ ->
			(let gn = gen_cd_name f in
			  cdl:=(gn,(f,l2))::!cdl;
			mess 2 ("Introducing the new complexity descriptor tau_"^gn^" over "^(cc(map string_of_struct l2)));
			  let cc = compl gftype
			  (map2 (fun x y -> join env x y) (params f) pars)
			  (map fst (params f)) (body f) in
			  cd_eqs:=(tau gn,cc)::!cd_eqs; cc)
			in subs_compl gftype env summants vars c

(**************************** compl_expr ***********************)
and compl_expr gftype env summants cgf x = 
if !printlevel>=4 then 
  print_string (make_string (2+2* !depth) ` ` ^ pp_exp x ^ " --> ");
let y = (foo x
where rec foo = function             (* cgf : current generating function *)
 
  Exp_constr(c,l)   -> Plus(Times(mu c,cgf),Pluss(map foo l))
| Call(f,pars)      -> let x = (if boundp f !functions then
			 compl_call gftype env summants f pars
			else (war ("function "^f^" is not defined.");
			 Pluss(map foo pars)))
		       and muf = mu f in
		       if muf=Gf_num(0) then x else Plus(Times(mu f,cgf),x)
| Exp_id id         -> Times(mu id,cgf)) in mess 4 (gts 0 y); y

(*************************** compl ****************************)
and compl gftype env summants = function

  Instructions(is) -> Pluss(map 
		   (fun i -> compl gftype env summants i) 
		   is)
| Cases(ids,cases) -> if for_all (fun id -> boundp id env) ids then
		       compl_cases gftype env summants ids 
				cases
		      else err "compl" "unknown variable in case" ids
| Casetype(id,cs)  -> if boundp id env then
			compl_casetype gftype env summants id cs
		      else err "compl" "unknown variable in casetype" id
| Expression(e)    -> compl_expr gftype env summants 
			(gf gftype env (tr_prod summants)) e
| Repeat(i,r)	   -> (match r with
	  RepNum(k)	-> Times(Gf_num k,compl gftype env summants i)
	| RepId(s)	-> if s<>"size" then err "compl" 
			  "repeat instruction other than size is not allowed" s
			   else Diff(compl gftype env summants i,1))
| Forall(x,y,i)	   -> if mem y !in_forall 
			then err "compl" "several forall on the same structure forbidden" y;
		      in_forall:=y::!in_forall;
		      K (if boundp x (!types@env) then err "compl" 
			"you can not use an assigned variable in forall" x
		      else if boundp y (!types@env) 
		      then (match get_def1 y (!types@env) with
			  Sequence(a,c) -> let tq = compl gftype
						((x,a)::env) [x] i
                                            and A = gf gftype env a in
                                let tp = compl_forall_sequence gftype tq A c in
				if eq_type (eval env (Ident(y,No)))
                                	(eval env (tr_prod summants)) then tp
				else subs_compl gftype env summants [y] tp
			| Cycle(a,c)	-> if eq_type (eval env (Ident(y,No)))
					     (eval env (tr_prod summants)) then
					    let tq = compl gftype
					     ((x,a)::env) [x] i
					    and A = gf gftype env a in
					     compl_forall_cycle gftype tq A c
					   else nyi "compl" "forall(cycle) for restricted type"
			| UCycle(a,c)	-> if eq_type (eval env (Ident(y,No)))
					     (eval env (tr_prod summants)) then
					    let tq = compl gftype
					     ((x,a)::env) [x] i
					    and A = gf gftype env a in
					     compl_forall_ucycle gftype tq A c
					   else nyi "compl" "forall(ucycle) for restricted type"
			| Set(a,c)	-> let tq = compl gftype
					     ((x,a)::env) [x] i
					   and A = gf gftype env a in
					    let tp = compl_forall_set gftype 
					      tq A c in gf gftype 
					       ((y,Ident(gts 5 tp,No))::env) 
						(tr_prod summants)
			| MultiSet(a,c)	-> let tq = compl gftype
					     ((x,a)::env) [x] i
					   and A = gf gftype env a in
					    let tp = compl_forall_multiset gftype 
					      tq A c in 
					subs_compl gftype env summants [y] tp
			| _		-> err "compl" "forall : neither a sequence nor a cycle nor a set nor a multiset" y)
		      else err "compl" "unknown variable" y)
			(in_forall:=tl !in_forall)
| Forone(x,y,i)	   -> if boundp x (!types@env) then err "compl" 
			"you can not use an assigned variable in forone" x
		      else if boundp y (!types@env)
		      then (match get_def1 y (!types@env) with
			  Sequence(a,c) -> let tq = compl gftype
						((x,a)::env) [x] i
                                            and A = gf gftype env a in
                                let tp = compl_forone_sequence gftype tq A c in
				if eq_type (eval env (Ident(y,No)))
                                	(eval env (tr_prod summants)) then tp
				else subs_compl gftype env summants [y] tp
			| Cycle(a,c)	-> if eq_type (eval env (Ident(y,No)))
					     (eval env (tr_prod summants)) then
					    let tq = compl gftype
					     ((x,a)::env) [x] i
					    and A = gf gftype env a in
					     compl_forone_cycle gftype tq A c
					   else nyi "compl" "forone(cycle) for restricted type"
			| Set(a,c)	-> let tq = compl gftype
					     ((x,a)::env) [x] i
					   and A = gf gftype env a in
					    let tp = compl_forone_set gftype 
					      tq A c in gf gftype 
					       ((y,Ident(gts 5 tp,No))::env) 
						(tr_prod summants)
			| _		-> err "compl" "forone : neither a sequence nor a cycle nor a set" y)
		      else err "compl" "unknown variable" y
| IfThenElse(c,t,e)-> (match c with
    Equal(Ident(x,_),y) -> if boundp x env then compl gftype 
			 env summants 
			 (Cases([x],[([y],t);([Pat_all],e)]))
	  	  	else err "compl" "unknown variable in if" x
  | Predicate(f,x)   -> let valx = eval env (Ident(x,No)) in
                         let pct = pcomplexity f true valx
                         and pcf = pcomplexity f false valx
			 in let then_env = subs_env x pct env
                            and else_env = subs_env x pcf env
                            in Pluss[compl gftype env summants
                                              (Expression(Call(f,[Exp_id x])));
                                    compl gftype then_env summants t;
			            compl gftype else_env summants e]
  | x                -> err "compl" "bad condition" x)
| Empty_instr	     -> Gf_num(0)

(********************** compl_cases ****************************)
and compl_cases gftype env summants ids = function
  []	-> Gf_num(0)
| c::x -> (match c with
	([Pat_all],i)-> compl gftype env summants i
      | ([pat],instr)   -> let st = struct_of_pat pat
			   and id = hd ids in 
			   (try (let l = filter env (Ident(id,No)) pat in
	  		    let newenv = ((id,st)::(l @ env)) 
			    and rest = 
		fst(subtract (eval env (Ident(id,No))) (eval (l@env) st)) in 
           Plus(compl gftype newenv 
		summants instr,
	        (if rest = Empty then Gf_num(0) 
		else compl_cases gftype (subs_env id rest env)
		       summants ids (c::x)))) with Failure "filter" ->
	 compl_cases gftype env summants ids x)
      | (pats,instr)    -> if list_length pats <> list_length ids then
			    err "compl_cases" "bad number of patterns" ids
			   else let sts = map struct_of_pat pats
				and l = flat(map2 (fun id p -> filter env
						     (Ident(id,No)) p) ids pats) in
				 let newenv = combine(ids,sts)@(l@env)
				 and restenv = env in
           Plus(compl gftype newenv 
		(replace ids sts summants) instr,
	        compl_cases gftype restenv
		 summants ids x))

(********************** compl_casetype *************************)
and compl_casetype gftype env summants id = function
  []    -> Gf_num(0)
| c::x -> (match c with
	([Pat_all],i)	-> compl gftype env summants i
      | ([Pat_id t],i)	-> let def = get_def2 id (!types@env) in
			    let f = (try filter_union t def with _ -> def) in
			  Plus(compl gftype ((id,Ident(t,No))::env) summants i,
				  compl_casetype gftype ((id,f)::env) summants
					id x)
      | x		-> pm "compl_casetype" x)
;;

(************************** complexity ************************)
let complexity gftype f = in_forall := [];
 if boundp f !functions then let fpars = params f in
  mess 4 ("Analyzing " ^ f ^ ":");
  K (compl gftype fpars (map fst fpars) (body f))
    (mess 4 "")
 else err "complexity" "unknown function" f;;


