#open "prelude";;
#open "output";;
#open "adl";;
#open "gf";;
#open "formulae";;

(******** Analyze of type of generating function ********
              (ordinary, exponential, ...)              *)

(* intersection of two gf types (ordinary, exponential, any) *)
let intersect_types s1 s2 = match (s1,s2) with
	  ("any",s)	-> s
	| (s,"any")	-> s
	| _		-> if s1=s2 then s1
			   else err "intersect_types" "different types" (s1,s2);;

(* finds the gf type of a list *)
let rec gftype s = function
	  []		-> s
	| (id,t)::y	-> intersect_types (gftype1 s t) (gftype s y)

and gftype1 s = function
	  Union(a,b)	-> intersect_types (gftype1 s a) (gftype1 s b)
	| Product(a,b)	-> intersect_types (gftype1 s a) (gftype1 s b)
	| Sequence(a,_)	-> gftype1 s a
	| Cycle(a,_)	-> gftype1 s a
	| Set(a,_)	-> gftype1 s a
	| MultiSet(a,_)	-> gftype1 s a
	| Atom(_,_,_)	-> intersect_types s "ordinary"
	| LAtom(_,_,_)	-> intersect_types s "exponential"
	| Ident(_,p)	-> if p=No then "any" 
				else intersect_types s "exponential"
	| _		-> "any";;

(* this function invokes get_def as long as x is an ident  *)
let rec get_def1 x env = match get_def x env with
  Ident(y,_)-> get_def1 y env
| z	    -> z ;;

let rec get_def2 x env = match get_def x env with
  Ident(y,_)-> get_def2 y env
| Atom(_,_,_) -> Ident(x,No)
| LAtom(_,_,_)-> Ident(x,No)
| z	    -> z ;;

(********************************** gf ****************************)
let rec number_of_diff = function
  []		-> 0
| Diff(_,k)::y	-> k + number_of_diff y
| _::y		-> number_of_diff y
;;

(* variables in vs do not vary  *)
let rec gf1 vs gftype env = foo where rec foo = function
  Union(a,b)	-> Plus(foo a,foo b)
| Empty		-> Gf_num(0)
| Epsilon	-> One
| Emptyset	-> One
| Product(a,b)  -> let l = map foo (untransform (Product(a,b))) in
		    (match number_of_diff l with
		       0 -> Timess(l)
		     | k -> Intg(Timess(l),k))
| OProduct(a,b) -> let l = map foo (untransform (Product(a,b))) in
		    (match number_of_diff l with
		       0 -> Timess(l)
		     | k -> Intg(Timess(l),k))
| Sequence(a,c)	-> gf_sequence gftype (foo a) c
| Cycle(a,c)	-> gf_cycle gftype (foo a) c
| UCycle(a,c)	-> gf_ucycle gftype (foo a) c
| Set(a,c)	-> gf_set gftype (foo a) c
| MultiSet(a,c)	-> gf_multiset gftype (foo a) c
| Ident(id,p)   -> if mem_assoc id vs then assoc id vs
	           else if boundp id env 
	 		        then foo (get_def id env)
	 		        else (match p with
			  No -> Gf_id(id)
			| _  -> Diff(Gf_id id,1))    (* Min | Max *)
| Atom(n,m,s)	-> gf_atom "ordinary" n m s
| LAtom(n,m,s)	-> gf_atom "exponential" n m s
| Mark(s)	-> Gf_par(s)
| Constr(p,a,c)	-> gf_constr p gftype (foo a) c
;;

let vz id = id ^ "0";;

let rec value_in_zero = function
  Union(a,b)    -> add_string (value_in_zero a) (value_in_zero b)
| Empty		-> "0"
| Epsilon	-> "1"
| Product(a,b)  -> mul_string (value_in_zero a) (value_in_zero b)
| Sequence(a,c) -> if lowest c = 0 then "1" else "0"
| Cycle(a,c)    -> "0"
| UCycle(a,c)   -> "0"
| Set(a,c)      -> if lowest c = 0 then "1" else "0"
| MultiSet(a,c) -> if lowest c = 0 then "1" else "0"
| Ident(id,_)	-> vz id
| Atom(n,m,_)   -> if n = 0 then son m else "0"
| LAtom(n,m,_)  -> if n = 0 then son m else "0"
| Mark(s)	-> "1"
| Emptyset	-> "1"
| x		-> pm "value_in_zero" x
;;

let gf gftype env = gf1 [] gftype env;;

(* the main counting function  *)
let count gftype t = 
 map (fun (id,x) -> (id,gf gftype [] x)) t;;

let without l x = filter_neg (fun y -> y=x) l;;

