#open "prelude";;
#open "output";;
#open "adl";;
#open "analyze";;
#open "types";;
#open "print";;
#open "same";;

let rec do_match y = foo where rec foo = function
  (Pat_id p,Exp_id e)	-> if boundp p !types then 
				if p=e then [] else failwith "do_match"
			   else [(p,Exp_id e)]
| (Pat_id p,e)		-> if boundp p !types then failwith "do_match"
			   else [(p,e)]
| (Pat_prod _,Exp_id _)	-> failwith "do_match"
| (Pat_prod(p,q),
Exp_constr("product",l))-> let l1=foo(p,hd l) and
			       l2=foo(q,if list_length l=2 then hd(tl l)
				else Exp_constr("product",tl l))
			   in l1@l2
| (Pat_all,e)		-> [(y,e)]
| xx			-> pm "do_match" xx
;;

let last_instr l = foo (rev l) where rec foo = function
  []   -> Empty_instr
| x::y -> if x=Empty_instr then foo y else x
;;

let split_instr l = foo (rev l) where rec foo = function
 []	-> ([],Empty_instr)
| x::y	-> if x=Empty_instr then foo y else (rev y,x)
;;

let rec expr_of_instr = function
  Expression(e) -> e
| Instructions(l)-> expr_of_instr (last_instr l)
| xx		-> err "expr_of_instr" "instruction shoud be an expression" xx
;;

let to_compose = ref [["toto";"tata"]];;
(* list of compositions to compute *)

(*   (cgf name,function name,bool,type)  *)
let pcomplexity_args = ref [("toto1",("toto",(true,Empty)))];;

let cdl = ref [("name999",("name",[Empty]))];; 
(* list of restricted procedures for which analysis has already begun :
	first element is the name of the c.d. (without "tau_")
	second element is the name of the general procedure
	third element is the list of restricted types *)

let gen_name l default f = (* generates a random name that is not in l *)
 let start=(if !printlevel<=2 then default		(* short names *)
			      else f)
 in if mem start l then gen_name_aux start l else start
;;

(* generates a random generating function name *)
let gen_gf_name f = (* DO NOT SKIP THE ARGUMENT f OTHERWISE THE BODY IS
			EVALUATED AT COMPILATION !!! *)
   gen_name ((map fst !types)@(map fst !pcomplexity_args)) "T" f
;;

let gen_cd_name f = (* generates a random complexity descriptor name *)
   gen_name ((map fst !cdl)@(map fst !functions)@(map snd !fnames)) "f" f
;;

let rec noc_aux = function
  f::(g::r)	-> let fg = f^"_of_"^g in
		      if not mem_assoc [f;g] !fnames then
			(let fg_name = gen_cd_name fg in
			 rap 2 ("Computing composition of "^f^" and "^g^" : "^fg_name);
			 fnames := ([f;g],fg_name)::!fnames;
			 if list_length !fnames >= 100 then
			 war ("the expansion process seems to loop: "^
				son(list_length !fnames)^
				" compositions have been defined");
			 to_compose := !to_compose @ [[f;g]]);
	(* append it at the end to compute compositions FIFO *)
		      let fg_name = assoc [f;g] !fnames in
		      if r <> [] then noc_aux (fg_name::r)
		      else fg_name
| [f]		-> f
| xx		-> err "noc_aux" "bad argument" xx
;;

(* returns the name of the composition if it already exists,
   otherwise creates a new name and returns it *)
let rec name_of_composition l = 
   if not mem_assoc l !fnames then 
	(let s = noc_aux l in
		if not mem_assoc l !fnames then fnames := (l,s)::!fnames;());
   assoc l !fnames
;;

(* l is a list of substitutions *)
let rec subs_in_expr l = function	(* expression -> expression *)
  Exp_id s		-> if mem_assoc s l then assoc s l else Exp_id s
| Exp_constr(c,ll)	-> Exp_constr(c,map (subs_in_expr l) ll)
| Call(f,ll)	-> Call((if mem_assoc f l then pp_exp(assoc f l) else f),
					map (subs_in_expr l) ll)
;;

let subs_in_cond l = function
  Equal(s,p) -> Equal(s,p)
| Predicate(s,t) -> if mem_assoc t l 
			then (match assoc t l with
		      Call(f,[Exp_id x])->Predicate(name_of_composition[s;f],x)
			| xx		-> Predicate(s,pp_exp xx))
			else Predicate(s,t)
;;

let rec subs_in_list l = function
  []	-> []
| x::y	-> (if mem_assoc x l then pp_exp(assoc x l) else x)::(subs_in_list l y)
;;

let rec subs_in_pat l p = match p with
  Pat_id(s)	-> if mem_assoc s l then Pat_id(pp_exp(assoc s l)) else p
| Pat_prod(q,r)	-> Pat_prod(subs_in_pat l q,subs_in_pat l r)
| xx		-> xx
;;

let rec subs_in_case l = function
 (patterns,instr) -> (map (subs_in_pat l) patterns,subs_in l instr)

and subs_in_cases l = function
  []	-> []
| c::y	-> (subs_in_case l c)::(subs_in_cases l y)

(* l : (string * expression) list *)
and subs_in l = function		(* AdlInstr -> AdlInstr *)
  Expression(e)	-> Expression(subs_in_expr l e)
| IfThenElse(c,t,e)	-> IfThenElse(subs_in_cond l c,subs_in l t,subs_in l e)
| Instructions(is)	-> Instructions(map (subs_in l) is)
| Cases(vars,cases)	-> Cases(subs_in_list l vars,subs_in_cases l cases)
| Casetype(s,cases)     -> Casetype(s,subs_in_cases l cases)
| Forall(a,b,i)		-> Forall(a,(if mem_assoc b l then pp_exp(assoc b l)
					else b),subs_in l i)
| Forone(a,b,i)		-> Forone(a,(if mem_assoc b l then pp_exp(assoc b l)
					else b),subs_in l i)
| Repeat(i,r)		-> Repeat(subs_in l i,r)
| Empty_instr		-> Empty_instr
;;

let rec comp_cases x y = function
  []		-> failwith "no pattern match case"
| ([p],i)::l	-> (try (let l=do_match y (p,x) in 
			subs_in ((y,x)::l) i)
		   with Failure "do_match" -> comp_cases x y l)
| (p,i)::l	-> err "comp_cases" "only one pattern allowed in function" p
;;

(* compute x y body -> body(y <- x) *)
let rec compute x y = function		(* AdlInstr -> AdlInstr *)
  Cases(_,cases) 	-> comp_cases x y cases
| Instructions(l)	-> compute x y (last_instr l)
			   (* let (ll,i) = split_instr l in
				Instructions(ll@[compute x y i]) *)
| Expression(e)	 	-> Expression(subs_in_expr [(y,x)] e)
| Empty_instr 		-> Empty_instr
| xx		 	-> pm "compute" xx
;;

let rec name_of_arg1 f = if mem_assoc f !functions then fst(hd(params f))
 else name_of_arg1(hd(rev(def_of_comp f)))
;;

(* ll:string list contains all bounded variables that can appear as
                    function arguments, in addition to constants *)
let rec compute_value_expr ll e = match e with
  Call(f,[x])		-> (let x' = expr_of_instr(compute_value_expr ll x) 
				in match x' with
	  Exp_id s	-> if boundp s !types 
			   then compute_value ll (compute x' 
				(name_of_arg1 f) (body f))
			   else if mem s ll then Expression(Call(f,[x']))
			   else user_err ("unable to compute \""^f^"\" of \""^
				s^"\"") ""
	| Call(g,_)	-> Expression(Call(f,[x']))
	| Exp_constr(_)	-> (try compute_value ll (compute x' 
				(fst(hd(params f))) (body f))
			    with Failure "no pattern match case" ->
			    user_err ("in function "^f^", no pattern for") 
				(pp_exp x'))
			)
| Exp_id(_)		-> Expression(e)
| Exp_constr(c,l)	-> Expression(Exp_constr(c,
			    map (fun y->expr_of_instr(compute_value_expr ll y))
				l))
| Call(f,l)		-> err "compute_value_expr"
				"several parameters in function composition" e
and compute_value ll = function
  Expression(e)	-> compute_value_expr ll e
| IfThenElse(c,t,e)	-> IfThenElse(c,compute_value ll t,compute_value ll e)
| Instructions(l)	-> Instructions(map (compute_value ll) l)
| xx			-> pm "compute_value" xx
;;

(* rewrite f(g(x)) where g : x -> bodyg *)
let rec rewrite ll f bodyg x = match bodyg with
  Cases(vars,cases)	-> if list_length(vars)<>1 or hd vars<>x then
			      err "rewrite" "bad variables in ``case''" vars;
			   Cases(vars,map (fun (p,i) -> (p,rewrite 
					  (flat(map vars_of_pattern p)) f i x))
					  cases)
| Expression(e) 	-> compute_value_expr (x::ll) (Call(f,[e]))
| Instructions(l)	-> let (is,i) = split_instr l in
				Instructions(is@[rewrite ll f i x])
| IfThenElse(c,t,e)	-> IfThenElse(c,rewrite ll f t x,rewrite ll f e x)
| xx			-> pm "rewrite" xx
;;

let rec abbrev_expr_aux l = function
  Exp_id s		-> Call(name_of_composition l,[Exp_id s])
| Call(g,[x])		-> abbrev_expr_aux (l@[g]) x
| xx			-> pm "abbrev_expr_aux" xx
;;

let rec abbrev_expr e = match e with
  Exp_id s		-> e
| Exp_constr(c,l)	-> Exp_constr(c,map abbrev_expr l)
| Call(f,[x])		-> abbrev_expr_aux [f] x
;;

let rec abbrev = function
  Cases(vars,cases)	-> Cases(vars,map (fun (p,i) -> (p,abbrev i)) cases)
| Expression(e)		-> Expression(abbrev_expr e)
| IfThenElse(c,t,e)	-> IfThenElse(c,abbrev t,abbrev e)
| Instructions(l)	-> Instructions(map abbrev l)
| xx			-> pm "abbrev" xx
;;

(* g : x -> y
   f : y -> z
   f o g : x -> z
   cost[ f(g(x)) ] = cost[ body(f) where y <- g(x) ] *)

(* compl_compo f g produces the body of f o g *)
let compl_compo f g =
   if output_type g <> input_type1 f then
      err "non compatible types for composition"
         (f^" : "^input_type1 f^" -> "^output_type f^", "^
	  g^" : "^input_type1 g^" -> "^output_type g) ();
   let bodyg = body g and x = fst(hd (params g)) in 
   let body_comp = abbrev(rewrite [] f bodyg x)
   and s = name_of_composition [f;g] in 
      if !printlevel>2 then
	(print_string ("\nfunction "^s^" ("^x^" : "^input_type1 g^
		") : "^output_type f^";\nbegin\n");
	print_string "   "; print_body 3 body_comp;
	print_string "end;\n\n");
      (s,(params g,(output_type f,body_comp)))
;;

(* replace function f by g in !functions *)
let rec replace_everywhere f g =
  functions := map (fun (s,(pars,(output_type,i))) -> 
	(s,(pars,(output_type,subs_in [(f,Exp_id g)] i)))) !functions;
  let (lp,ln) = partition (mem f) !to_compose
  in to_compose := ln;
  map (fun x -> let w = assoc x !fnames in
		let x' = subs_in_list [(f,Exp_id g)] x in
		fnames := without !fnames (x,w);
		if mem_assoc x' !fnames then
		   replace_everywhere w (assoc x' !fnames)
		else
		   (to_compose := x'::!to_compose;
		   fnames := (x',w)::!fnames);())
	lp;
  fnames := map (fun (l,s) -> if s=f then (l,g) else (l,s)) !fnames
;;

let rec reduce_compositions () = match !to_compose with
  []		-> ()
| [f;g]::l	-> (let fg_comp = compl_compo f g in match fg_comp with
		     (fg,(pars,(output_type,body_fg))) ->
		   let normal_fg = normal2 fg_comp
		   and input_type_fg = input_type_aux pars in
		   (try let ff=find (same_function2 (input_type_fg,output_type,
			normal_fg)) (map fst !functions) in
			replace_everywhere fg ff;
			rap 2 ("Function "^fg^" is the same as "^ff)
		   with Failure "find" -> functions := fg_comp::!functions);
		   to_compose := without !to_compose [f;g];
		   reduce_compositions())
;;
