#open "prelude";;
#open "output";;
#open "maple";;

type gf = Gf_id of string	(* generating function *)
	| Gf_par of string	(* parameter *)
	| Gf_num of int
	| Plus of gf * gf
	| Pluss of gf list
	| Minus of gf * gf
	| UMinus of gf
	| OneMinus of gf
	| OnePlus of gf
	| Times of gf * gf
	| Timess of gf list
	| Ratio of gf * gf
	| NRatio of gf * int
	| Power of gf * int
	| Polya of gf * int	(* if int = -1  ==> all else card=int *)
	| MPolya of gf * int	(* if int = -1  ==> all else card=int *)
	| Texp of gf * int	(* Texp(x,k) = 1 + x + x^2/2 + ... + x^k/k! *)
	| Tder of gf * int	(* Tder(x,k) = 1 + 2 x + ... + k x^{k-1} *)
	| Fact of int
	| Z
	| One
	| Op of string * gf	(* Exp, L, Q, ... *)
	| Intg of gf * int
	| Diff of gf * int
;;

let rec cost_in_zero = function
  Gf_id(s)	-> if last_char s = `)` then "subs(z=0,"^s^")" (* wrong *)
		   else s^"0"
| Gf_par(s)	-> s
| Gf_num(k)	-> son k
| Plus(x,y)	-> "("^cost_in_zero x^"+"^cost_in_zero y^")"
| Pluss([])	-> "0"
| Pluss(x::y)	-> cost_in_zero (Plus(x,Pluss(y)))
| Minus(x,y)	-> "("^cost_in_zero x^"-"^cost_in_zero y^")"
| UMinus(x)     -> "-"^cost_in_zero x
| OneMinus(x)   -> cost_in_zero (Minus(One,x))
| OnePlus(x)    -> cost_in_zero (Plus(One,x))
| Times(x,y)	-> "("^cost_in_zero x^"*"^cost_in_zero y^")"
| Timess([])    -> "1"
| Timess(x::r)  -> cost_in_zero (Times(x,Timess(r)))
| Ratio(x,y)    -> "("^cost_in_zero x^"/"^cost_in_zero y^")"
| NRatio(x,k)   -> cost_in_zero (Ratio(x,Gf_num(k)))
| Power(x,k)    -> cost_in_zero x^"^"^son k
| Polya(x,-1)   -> "1"
| Polya(x,k)    -> "Polya(" ^ cost_in_zero x ^ "," ^ son k ^ ")"
| MPolya(x,-1)  -> "1"
| MPolya(x,k)   -> "MPolya(" ^ cost_in_zero x ^ "," ^ son k ^ ")"
| Texp(x,k)     -> "texp(" ^ cost_in_zero x ^ "," ^ son k ^ ")"
| Tder(x,k)	-> "tder(" ^ cost_in_zero x ^ "," ^ son k ^ ")"
| Fact(k)       -> son k ^ "!"
| Z             -> "z"
| One           -> "1"
| Op(s,x)       -> if mem s ["Q";"L"] then "0"
			(* in Q(g(z)) or L(g(z)), necessarily g(0)=0 *)
		   else s ^ "(" ^ cost_in_zero x ^ ")"
| Intg(x,k)     -> "0"
| Diff(x,k)     -> "diff(" ^ cost_in_zero x ^ ",z$" ^ son k ^ ")"
;;

(* d=0 : no need of parentheses
   d>=1 : parentheses for unary - 
   d>=2 : parentheses for + or binary -
   d>=3 : parentheses for /
   d>=4 : parentheses for *
   d>=5 : parentheses for ^
*)

let rec gts d = function

  Gf_id(s)	-> if last_char s = `)` then s else (maple_name s^"(z)")
| Gf_par(s)	-> s
| Gf_num(k)	-> son k
| Plus(x,y)	-> if d>=2 then "(" ^ gts 0 x ^ "+" ^ gts 1 y ^ ")"
		   else gts d x ^ "+" ^ gts 1 y
| Pluss([])	-> "0"
| Pluss([x])	-> gts d x
| Pluss([x;y])	-> gts d (Plus(x,y))
| Pluss(x::r)	-> gts d (Plus(x,Pluss(r)))
| Minus(x,y)	-> if d>=2 then "(" ^ gts 0 x ^ "-" ^ gts 2 y ^ ")"
		   else gts d x ^ "-" ^ gts 2 y
| UMinus(x)	-> if d>=1 then "(-" ^ gts 5 x ^ ")" else "-" ^ gts 5 x
| OneMinus(x)	-> gts d (Minus(One,x))
| OnePlus(x)	-> gts d (Plus(One,x))
| Times(x,y)	-> if d>=4 then "(" ^ gts 2 x ^ "*" ^ gts 2 y ^ ")"
		   else gts (max d 2) x ^ "*" ^ gts (max d 2) y
| Timess([])	-> "1"
| Timess([x])	-> gts d x
| Timess([x;y])	-> gts d (Times(x,y))
| Timess(x::r)	-> gts d (Times(x,Timess(r)))
| Ratio(x,y)	-> if d>=3 then "(" ^ gts 2 x ^ "/" ^ gts 4 y ^ ")"
		   else gts 2 x ^ "/" ^ gts 4 y
| NRatio(x,k)	-> gts d (Ratio(x,Gf_num(k)))
| Power(x,k)	-> if d>=5 then "(" ^ gts 5 x ^ "^" ^ son k ^ ")"
		   else gts 5 x ^ "^" ^ son k
| Polya(x,-1)	-> "PPolya(" ^ gts 0 x ^ ")"
| Polya(x,k)	-> "Polya(" ^ gts 0 x ^ "," ^ son k ^ ")"
| MPolya(x,-1)	-> "MP(" ^ gts 0 x ^ ")"
| MPolya(x,k)	-> "MPolya(" ^ gts 0 x ^ "," ^ son k ^ ")"
| Texp(x,k)	-> "texp(" ^ gts 0 x ^ "," ^ son k ^ ")"
| Tder(x,k)	-> "tder(" ^ gts 0 x ^ "," ^ son k ^ ")"
| Fact(k)	-> son k ^ "!"
| Z		-> "z"
| One		-> "1"
| Op(s,x)	-> s ^ "(" ^ gts 0 x ^ ")"
| Intg(x,k)	-> "intg(" ^ gts 0 x ^ ",z," ^ son k ^ ")"
| Diff(x,k)	-> "diff(" ^ gts 0 x ^ ",z$" ^ son k ^ ")"
;; 
