#open "sys";;
#open "prelude";;
#open "output";; (* for printlevel *)

let maple_loaded = ref false ;;		(* has Maple been initialized ? *)
let maple_in_chan = ref std_out
and maple_out_chan = ref std_in
and maple_buf = ref "xxxxx"
and luo_dir = ref ""
;;

let input_string chan n = input chan !maple_buf 0 n;sub_string !maple_buf 0 n;;

let mapleget = function			(* to get one char of Maple *)
  () -> input_string !maple_out_chan 1;;

let maple_get_line = function		(* to get one line of Maple *)
  () -> input_line !maple_out_chan;;

let rec mapleput s = let l = string_length s in
if l < 150 then 	(* 250 is too fast for Maple *)
	(output_string !maple_in_chan (s ^ "\n");flush !maple_in_chan)
else (
      output_string !maple_in_chan (sub_string s 0 150 ^ "\\\n");
      flush !maple_in_chan;
      mapleput (sub_string s 150 (l-150)));;

let is_prefix s1 s2 = mem (compare_strings s1 s2) [0;-2];;

(* this function traps Maple's errors and other messages *)
let rec my_echo_string echop s =
if is_prefix "Error, (in" s then 
 (print_string s; mapleflush true; failwith "maple")
else if echop & not (is_prefix "lprint(`#$#`);" s)
then (print_string s; print_newline())

(* this function recognizes when Maple has finished and flushes its buffer *)
and mapleflush = function
  echop  -> let s = maple_get_line () in
	(
	if !printlevel>=10 then (print_string ("got line: "^s); 
		print_newline(); flush std_out);
	 if not is_prefix "#$" s
	  then (my_echo_string echop s; 
		mapleflush echop))

and mapleflush0 echop = mapleput "lprint(`#$#`);"; 
	mapleflush echop;;

let maple_eval = function
   s -> if !maple_loaded then
	 (mapleput s;
	  mapleput "lprint(`#$#`);";
	  (* maple_get_line ();	*)	(* to skip input line : bug in V.3 *)
	  mapleflush true);;
    
let maple_eval_quiet = function
   s -> if !maple_loaded then
	 (mapleput s; mapleflush0 false);;
    
let mapleinit = function
  () -> let opty = getenv "OPTY"
	and ipty = getenv "IPTY" in
	 if opty<>"" & ipty<>"" & !maple_loaded=false then
	  (maple_in_chan := open_out opty;
	   maple_out_chan := open_in ipty;
	   maple_loaded := true;
	   maple_eval_quiet ("read `" ^ !luo_dir ^ "/Maple/Solve.m`;");
	   maple_eval_quiet ("read `" ^ !luo_dir ^ "/Maple/distribution.m`;");
	   maple_eval_quiet ("libname:=`" ^ !luo_dir ^ "/Maple`,libname:");
	   maple_eval_quiet "readlib(equivalent):";
           (* the following line avoids that strange signums appear in O(...) *)
           maple_eval_quiet "assume(_Xasy>100):"
	  );;

let maplequit = function
  () -> mapleput "quit();";
        input_string !maple_out_chan 5;
	print_string (input_line !maple_out_chan);
        close_in !maple_out_chan;
        close_out !maple_in_chan;;
	
let rec get_maple_string = function
  () -> let c = mapleget() in
	  if c = "#" then ""
          else if (int_of_char (nth_char c 0) < 32) or (c = "\\") then get_maple_string ()
	  else c ^ get_maple_string ();;

let respond_to_caml = function
  () -> mapleput "lprint(\");";
	mapleput "lprint(`#`);";
	let s = get_maple_string ()
        in
	if !printlevel>=10 then (print_string ("got line: "^s); 
		print_newline(); flush std_out);
	s;;

let to_maple = function
  s -> if !maple_loaded then
	 (let s1 = if mem (last_char s) [`;`;`:`] then s else s^":" in
	   mapleput s1; 
	   mapleflush0 false;		(* without echo *)
       	   let s2 = respond_to_caml () in
	if !printlevel>=10 then (print_string ("extracted "^s2); 
		print_newline(); flush std_out);
	   s2)
       else s;;

let maple_names = ref [("toto","tata")];;

let maple_name s = 
   if mem_assoc s !maple_names then assoc s !maple_names
   else
      (let t = if to_maple ("is_a_maple_keyword("^s^")")="false"
	then "`"^s^"`" else "` "^s^"`" in
	maple_names := (s,t)::!maple_names;
	t);;

