#open "prelude";;
#open "output";;
#open "adl";;
#open "comp";;
#open "types";;
#open "analyze";;

       (*   computation of conditional generating functions   *)

(* substitues the newest definition of id by val in env *)
let subs_env id val env = (id,val)::(foo env) where rec foo = function
  []	-> []
| p::q	-> if fst p = id then foo q else p::(foo q);;

(*************************** replace1 **************************)
let replace1 (id,struct) = foo where rec foo = function
  Ident(s,p)	-> if s=id then struct else Ident(s,p)
| Union(x,y)	-> Union(foo x,foo y)
| Product(x,y)	-> Product(foo x,foo y)
| Sequence(x,c)	-> Sequence(foo x,c)
| Cycle(x,c)	-> Cycle(foo x,c)
| Set(x,c)	-> Set(foo x,c)
| x		-> x;;

let rec cond_gf f bool st = function
  [] -> failwith "no"
| (a,(b,(c,d)))::y -> if b=f & c=bool & d=st then Ident(a,No)
                  else cond_gf f bool st y;;

let gen_cgf t bool f =
(* generates a name for the sub-structure of t such that f is true *)
 let pre = t^(if bool then "_" else "_not_")^f 
 in gen_gf_name pre
;;

(************************ pcompl_call **********************)
let rec pcompl_call bool env current_type f pars = match pars with
  []        -> err "pcompl_call" "empty list of parameters for function " f
| [Exp_id y]-> let parf = params f in (* parf : type_decl list *)
	       if list_length parf<>1 
		then err "pcompl_call" "bad number of parameters" parf
	        else let pc = pcomplexity f bool (eval env (Ident(y,No))) in
		 eval ((y,pc)::env) current_type
| [Call(g,l)]-> let s = name_of_composition [f;g] in
			reduce_compositions();
			pcompl_call bool env current_type s l
| x         -> err "pcompl_call" "several parameters not yet implemented" x
  
(**************************** pcompl_expr ***********************)
and pcompl_expr bool env current_type = (foo
where rec foo = function
 
  Exp_constr(c,l)   -> err "pcompl_expr" "result must be boolean, not" c
| Call(f,pars)      -> pcompl_call bool env current_type f pars
| Exp_id id         -> if bool_of_string id = bool then eval env current_type
                       else Empty)

(*************************** pcompl ****************************)
and pcompl bool env current_type = function

  Instructions(l)  -> pcompl bool env current_type (last_instr l)
| Cases(ids,cases)  -> if for_all (fun id -> boundp id env) ids then
			pcompl_cases bool env current_type ids cases
		       else err "pcompl" "unknown variable in case" ids
| Expression(e)    -> pcompl_expr bool env current_type e
| IfThenElse(c,t,e)-> (match c with
    Equal(Ident(x,_),y) -> if boundp x env 
			then pcompl bool env current_type 
			 (Cases([x],[([y],t);([Pat_all],e)]))
	  	  	else err "pcompl" "unknown variable in if" x
  | Predicate(f,x)   -> let valx = eval env (Ident(x,No))
                        in let pct = pcomplexity f true valx
                           and pcf = pcomplexity f false valx
                      in Union(pcompl bool (subs_env x pct env) current_type t,
                               pcompl bool (subs_env x pcf env) current_type e)
  | x                -> err "pcompl" "bad condition" x)
| Empty_instr	     -> err "pcompl" 
			    "empty instruction : no boolean value returned" ()
| Forall(x,y,i)	     -> (match get_def1 y (!types@env) with
			  MultiSet(a,All) -> if bool then MultiSet(pcompl true 
					      ((x,a)::env) (Ident(x,No)) i,All)
			    else Product(MultiSet(pcompl true ((x,a)::env) 
						(Ident(x,No)) i,All),
					 MultiSet(pcompl false ((x,a)::env)
						(Ident(x,No)) i,(Ge(All,1))))
			)
| Forone(x,y,i)	     -> (match get_def1 y (!types@env) with
				(* Rule 5 of [Zimmermann92a] *)
			  Sequence(a,All) -> if not bool then 
						Sequence(pcompl false
					      ((x,a)::env) (Ident(x,No)) i,All)
			    else Product(Sequence(pcompl false
					((x,a)::env) (Ident(x,No)) i,All),
					Product(pcompl true ((x,a)::env) 
						(Ident(x,No)) i,
					 Sequence(a,All)))
			)
| x		     -> pm "pcompl" x

(********************** pcompl_cases ****************************)
and pcompl_cases bool env current_type ids = function
  []   -> Empty
| c::x -> (match c with
	([Pat_all],i)-> pcompl bool env current_type i
      | ([pat],instr)  ->let id = hd ids in
			  try (let l = filter env (Ident(id,No)) pat
	  		  and st = struct_of_pat pat in 
	  		   let newenv = subs_env id st (l @ env) 
			   and restenv = subs_env id
	(fst(subtract (eval env (Ident(id,No))) (eval (l@env) st))) env in
           Union(pcompl bool newenv 
		 	(replace1 (id,st) current_type) instr,
	   	 pcompl_cases bool restenv
			current_type ids x))
	      with Failure "filter" -> pcompl_cases bool env current_type ids x
      | xx -> err "pcompl_cases" "several patterns" xx)

(************************** pcomplexity ************************)
(* computes the conditional generating function of objects x in st such
   that fname(x) = bool *)
and pcomplexity fname bool st =
let t = (match st with (Ident(s,_)) -> s | _ -> "type") in
 let l=filter_pos (fun x -> fst x = t & fst(snd x)= fname) !pcomplexity_args in
 if l<>[] then (if fst(snd(snd(hd l)))=bool then st else Empty)
 else try (cond_gf fname bool st !pcomplexity_args)
 with Failure "no" -> let nn = gen_cgf t bool fname in
 (pcomplexity_args:=(nn,(fname,(bool,st)))::!pcomplexity_args;
  mess 2 ("Introducing the new type "^nn^" for which function "^fname^" returns "^string_of_bool bool);
  try (let res = match assoc fname !functions with
	(fpars,(ftype,body))->if list_length fpars <> 1 then err "pcomplexity" 
			"several parameters : not yet implemented" ()
                              else if ftype <> "boolean" then
				err "pcomplexity" "not a predicate" fname
			      else let (x1,type1) = hd fpars in
			simplify (pcompl bool [(x1,st)] (Ident(x1,No)) body)
     in types:=(nn,res)::!types;Ident(nn,No))
 with Failure "find" -> user_err "undefined function" fname);;
