#open "prelude";;
#open "output";;
#open "adl";;

let string_of_tdl tdl =
   let l1 = map (fun (s,st) -> s^" : "^string_of_struct st) tdl
   in end_it_list (prefix ^) "" l1
;;

let indent n = print_string (make_string n ` `);;

let rec string_of_pat = function
  Pat_id s	-> s
| Pat_all	-> "_"
| Pat_epsilon	-> "()"
| Pat_emptyset	-> "{}"
| Pat_prod(p,q) -> "("^string_of_pat p^","^string_of_pat q^")"
;;

let pp_cond = function
  Equal(s,p)	-> print_string (string_of_struct s^" = "^string_of_pat p)
| Predicate(s,t)-> print_string (s ^ "(" ^ t ^ ")")
;;

let rec print_body ind = function
  Cases(vars,cases) -> output_line std_out ("case "^cc vars^" of");
			map (fun (p,i) -> indent (ind+3); print_string
					  (cc (map string_of_pat p)^" : ");
					  print_body (ind+3) i;
					  output_line std_out ";") cases;
		       indent ind; output_line std_out "end;"
| Expression(e)	    -> print_string (pp_exp e)
| IfThenElse(c,t,e) -> print_string "if "; pp_cond c;print_string " then ";
			print_body ind t; print_string " else ";
			print_body ind e
| Instructions(l)   -> output_line std_out "begin";
			map (fun i->print_body(ind+3) i;print_string "; ") l;
			output_line std_out "end;"
| xx -> pm "print_body" xx
;;

let print_function f =
   let (tdl,(typ,instrs)) = get_def f !functions
   in output_line std_out ("function "^f^"("^string_of_tdl tdl^") : "^typ^";");
      output_line std_out "begin";
      indent 3; print_body 3 instrs;
      output_line std_out "end;"
;;
   
