#open "sys";;
#open "prelude";;
#open "output";;
#open "maple";;
#open "adl";;
#open "gf";;
#open "analyze";;
#open "types";;
#open "comp";;
#open "ana";;
#open "adl_parser";;
#open "cite";;
#open "check";;
#open "distrib";;

let start_maple_time = ref 0.0
and solve_time = ref 0.0
and equiv_time = ref 0.0
;;

let maple_time () = float_of_string (to_maple "time()");;

let equivalent n eq f m k = (* f is the generating function, eg a(z)*b(z) *)
			    (* eq is the name for the equivalent, eg a_b *)
let ord = string_of_int !order in
maple_eval_quiet (eq^"_n:=subsEquiv("^f^",z,"^ord^","^m^","^k^"):");
if !printlevel >= n then maple_eval ("my_printdev("^eq^"_n,"^ord^");");
if m<>"1" then rap n ("for n mod "^m^" = "^k^", and 0 otherwise.");;

let divide f teq = let s = "tau_"^f^"_n" in
   maple_eval ("my_dividev("^s^","^teq^"_n,"^son !order^",'av_"^s^"');");;

let do_evalf s = rap 1 "Floating point evaluation:\n\n";
 maple_eval ("luo_evalf(" ^ s ^ "_n," ^ son !order ^ ");");;

let eq_of_pair (x,y) = x^"="^y ;;

(* gives an equation to Maple, and returns its name *)
let eq_names = ref ["eq0"];;

let define_equation s = let n = gen_name !eq_names "eq" "eq" in
	maple_eval_quiet (n ^ ":=" ^ s ^ ":");
	if !printlevel>=2 then maple_eval "lprint(``,\");";
	eq_names := n::!eq_names;
	n
;;

let eq_of_pair2 (x,y) = 
	define_equation((gts 0 (Gf_id x)) ^ "=" ^ (gts 0 y));;

let showone s = rap 2 ("  " ^ s ^ ",");;
let showone2 s = maple_eval ("lprint(``," ^ s ^ ");");;

let indic s = rap 2 (">>> " ^ s ^ " ...\n");;

let gtype = ref "";;

(* m: mark	taus,s: struct		k: order *)
let compute_moment m taus s k =
 if k<=0 then err "moment" "last argument must be >= 1" k
 else rap 1 ("Moment of order "^son k^" associated to mark " ^ m ^ " in " ^ taus ^ " is:");
 let ev = string_of_bool !evalf in
maple_eval ("moment("^m^","^maple_name taus^","^maple_name s^","^son k^","^son !order^","^ev^"):");;

let compute_variance m taus s =
 rap 1 ("Variance associated to mark " ^ m ^ " in " ^ taus ^ " is:");
 let ev = string_of_bool !evalf in
maple_eval ("variance("^m^","^maple_name taus^","^maple_name s^","^son !order^","^ev^");");; 

let init_analysis () = cdl:=[]; cd_eqs:=[]; pcomplexity_args:=[];
   id_not_def:=["true";"false";"product"];
   fnames:=[]; to_compose:=[];	(* for function composition *)
   to_cite:=[]
;;

let cd_equations () =
   let b = map (fun (f,_) -> (tau f,complexity !gtype f)) !functions
   in let new_names = map snd !fnames
   in let new_fcts = filter_pos (fun x -> mem (fst x) new_names) !functions
   in b @ (map (fun (f,_) -> (tau f,complexity !gtype f)) new_fcts) @ !cd_eqs 
;;

let load_adl s = match adl_parse s with
  File(t,f,m) -> init_analysis(); types:=t; functions:=f; measures:=m;
  gtype := gftype "any" !types; ()
;;

let alas_and_solver () =
 rep "\\section{Algebraic analysis}\n";
 rep "{\\verbsize\\begin{verbatim}";
 check_all ();
 if !maple_loaded then maple_eval_quiet "`luo/unassign`():";
 indic "ALGEBRAIC ANALYZER";
 rap 2 ("Generating functions are " ^ !gtype ^ ".\n");
 let b = cd_equations () and a = count !gtype !types
 in rap 2 "Counting generating functions:";
    let a1 = map eq_of_pair2 a
    in rap 2 "";
   if b<>[] then rap 2 "Complexity descriptors:";
   let b1 = map eq_of_pair2 b in
   rep "\\end{verbatim}}\n";
   if b<>[] then rap 2 "";
   rep "\\section{Solving equations}\n";
   rep "{\\verbsize\\begin{verbatim}";
   let eqs = "{" ^ cc a1 ^ "},{" ^ cc b1 ^ "},[],{}"
    in if !maple_loaded & !do_solve then
	  (maple_eval_quiet ("gftype:=`"^ !gtype^"`:");
	   maple_eval_quiet ("marks:={"^end_it_list (fun s t -> s^","^t) "" 
		!marks^"}:");
	   maple_eval ("luo1(" ^ eqs ^ "):");
	   rep "\\end{verbatim}}\n";
	   biblio())
;;

let maxvar s = match words(to_maple("maxvar0("^s^",z)")) with
  [k;m] -> (k,m)
| _ -> snh 1
;;

let asympt_type s = 
 if !gtype = "ordinary" then rap 1 ("Number of " ^ s ^ " of size n is:")
 else if !gtype = "exponential" then 
	rap 1 ("Number of "^s^" of size n is n! times:")
 else err "Analyze" "unknown gf type" !gtype;
 let Sofz = (maple_name s)^"(z)" in
 let (k,m) = maxvar Sofz in
    print_newline();
    equivalent 0 s Sofz m k; 
    if !evalf & !printlevel>0 then do_evalf s
;;

let asympt_function f = let l = params f in
 let l1 = map (function (_,Ident(x,_))->x | z->err "Analyze" "bad type" z) l in
	     let N = end_it_list (fun x y -> x^"*"^y) "1" 
			(map (fun x -> (maple_name x)^"(z)") l1)
    	     and eqN = end_it_list (fun x y -> x^"_"^y) "" l1
	     and tau_f = tau f in
	      let (k,m) = maxvar(N^"+"^tau_f^"(z)") in
rap 2 ("Function "^f^":");
if !gtype = "ordinary" then
 (rap 2 ("Number of inputs of "^f^" of size n is:\n");
  equivalent 2 eqN N m k;
  if m<>"1" then rap 2 "";
  rap 2 ("Total cost for "^f^" on all inputs of size n is:\n");
  equivalent 2 tau_f (tau_f^"(z)") m k)
else if !gtype = "exponential" then
 (rap 2 ("Number of inputs of "^f^" of size n is n! times:\n");
  equivalent 2 eqN N m k;
  rap 2 ("Total cost of "^f^" on all inputs of size n is n! times:\n");
  equivalent 2 tau_f (tau_f^"(z)") m k)
else err "Analyze" "unknown gf type" !gtype; 
rap 1 ("Average cost for "^f^" on random inputs of size n is:"); 
print_newline();
divide f eqN;
if m<>"1" then rap 1 ("for n mod "^m^" = "^k^", and 0 otherwise.");
if !evalf & !printlevel>0 then do_evalf ("av_"^tau_f)
;;

let asympt s =
 if mem_assoc s !types then asympt_type s
 else if mem_assoc s !functions then asympt_function s
 else failwith ("no data type nor function ``"^s^"'' in current Adl program");
 biblio()
;;

let check_type s = if not mem_assoc s !types then
	failwith (s^" is not a data type of current Adl program");;

let moment m s k = check_type s; compute_moment m s s k; biblio();;

let variance m s = check_type s; compute_variance m s s; biblio();;

let distribution m s = check_type s; compute_distr m s; biblio();;

let base_adl_name file = file^".adl";;
let find_adl_file file = file^".adl";;

let trunc_time n = string_of_float (add_float n 0.05);;

let analyze1 file = let s = find_adl_file file in
   types:=[]; marks:=[];
(*   ("Analyze <<" ^ (cat s) ^ ">>;;\n"); *)
   if !timings then
       (print_string "Algebraic Analyzer: ";
	print_string "s  Solver: ";
	print_string (trunc_time !solve_time);
	print_string "s  Analytic Analyzer: ";
	print_string (trunc_time !equiv_time);
	print_string "s\n");;

let analyze file =
 report_mode := false;
 analyze1 file;;

let screenwidth n = maple_eval_quiet ("screenwidth:="^(string_of_int n)^":");;

let report2 file title = let ban = base_adl_name file
		   and oldprintlevel = !printlevel
		   and mode = if !chapter then "chapter" else "article"
in
    screenwidth 99; printlevel:=2;
    report_mode := true;
    anafile := open_out (ban ^ ".ana");
    (try analyze1 file with Failure _ -> ());
    close_out !anafile;
    screenwidth 80; printlevel:=oldprintlevel;
    output_line std_out ("Generating "^file^".tex ...");
    comline (!luo_dir^"/bin/report "^ban^" \'"^title^"\' "^mode)
;;

let report file = report2 file file;; (* title = file *)

let help s = 
 if s="" then cat (!luo_dir^"/Help/main")
 else let f = !luo_dir^"/Help/"^s in
  try cat f with Sys_error "No such file or directory" -> 
	print_string ("There is no help for the command: "^s^"\n");
	print_string "For more help, type ? followed by a carriage return.\n"
;;
