#open "output";;
#open "adl";;
#open "types";;

let assoc_list = ref [("distribute","v1")]
and excluded_names = ref ["danger"]
and last_number = ref 0;;

(* generates a name beginning with f that is not in <names> *)
let gen_name_aux f names = foo (!last_number+1) where rec foo n =
   let s = f^son n in if mem s names then foo (n+1) 
   else (last_number:=n; s);;

(* if s may be a free variable *)
let normal_free s = if boundp s !types then s	(* data type *)
		   else if mem_assoc s !assoc_list 
			then assoc s !assoc_list
			else (* free variable *)
			let news = gen_name_aux "v" !excluded_names
			in assoc_list:=(s,news)::!assoc_list; news
;;

(* if s can not be a free variable *)
let normal_non_free s = 
	if mem_assoc s !assoc_list then assoc s !assoc_list else s
;;

let rec normal_pat p = match p with
  Pat_id(s)	-> Pat_id(normal_free s)
| Pat_prod(q,r)	-> Pat_prod(normal_pat q,normal_pat r)
| _		-> p
;;

let normal_cond = function
  Equal(s,p)	-> Equal(s,normal_pat p)
| Predicate(f,x)-> Predicate(normal_non_free f,normal_non_free x)
;;

let rec normal_case = function
  (pats,instr) -> (map normal_pat pats,normal_instr instr)

and normal_cases = function
  []		-> []
| c::y		-> (normal_case c)::(normal_cases y)

and normal_instr = function
  Instructions(is) -> if list_length(is)=1 then normal_instr (hd is)
			else Instructions(map normal_instr is)
| Cases(vars,cases)-> Cases(map normal_non_free vars,normal_cases cases)
| Expression(e)	   -> Expression(normal_expr e)
| Casetype(s,cases)->Casetype(normal_non_free s,normal_cases cases)
| IfThenElse(c,t,e)->IfThenElse(normal_cond c,normal_instr t,normal_instr e)
| Forall(s,t,i)	   -> Forall(normal_free s,normal_non_free t,normal_instr i)
| Forone(s,t,i)    -> Forone(normal_free s,normal_non_free t,normal_instr i)
| Repeat(i,r)	   -> Repeat(normal_instr i,r)
| Empty_instr	   -> Empty_instr

and normal_expr = function
  Exp_id(s)	-> if mem_assoc s !assoc_list then Exp_id(assoc s !assoc_list)
		   else Exp_id(s)
| Exp_constr(t,l)->Exp_constr(t,map normal_expr l)
| Call(f,l)	->Call((if mem_assoc f !assoc_list then assoc f !assoc_list
			else f),map normal_expr l)
;;

(* replaces all free variables by standard names v0, v1, ... *)
let normal2 (f,(pars,(out_type,instr))) = last_number:=0;
   excluded_names := (map fst !types)@(map fst !functions);
      let newf = gen_name_aux "v" !excluded_names
      and newpars = map (fun p->(gen_name_aux "v" !excluded_names,snd p)) pars
      in assoc_list := ((f,newf)::(combine(map fst pars,map fst newpars)));
      (newf,newpars,out_type,normal_instr instr)
;;

(* replaces all free variables by standard names v0, v1, ... *)
let normal f = normal2(f,get_def f !functions);;

(* for use in reduce_compositions *)
let same_function2 (input_type_f,output_type_f,normal_f) g =
   if input_type_f <> input_type g then false
   else if output_type_f <> output_type g then false
   else normal_f = normal g
;;
	
(* for toplevel *)
let same_function f = same_function2 (input_type f,output_type f,normal f)
;;

