#open "prelude";;
#open "output";;
#open "adl";;
#open "analyze";;
#open "gf";;

(*************************** eval ****************************)
(* env   : list of variable's types *)
let eval env = foo
 where rec foo = function
  Ident(x,p)    -> if boundp x env then foo(get_def x env) else Ident(x,p)
| Epsilon   	-> Epsilon
| Emptyset	-> Emptyset
| Empty		-> Empty
| Union(x,y)    -> Union(foo x,foo y)
| Product(x,y)  -> Product(foo x,foo y)
| OProduct(x,y) -> OProduct(foo x,foo y)
| Sequence(x,c) -> Sequence(foo x,c)
| Set(x,c)	-> Set(foo x,c)
| MultiSet(x,c)	-> MultiSet(foo x,c)
| Cycle(x,c)	-> Cycle(foo x,c)
| xx            -> pm "eval" xx;;
  
(************************ atomp ******************************)
let atomp s = try (match get_def s !types with
  Atom(_,_,_)	-> true
| LAtom(_,_,_)	-> true
| _ 	  	-> false) 
with Failure _ -> err "atomp" "unknown type" s;;

let atom_size s = match get_def s !types with
  Atom(k,_,_)	-> k
| LAtom(k,_,_)	-> k
| _		-> err "atom_size" "not an atom" s;;

(************************ simplify ***************************)
let rec simplify = function
  Union(x,y)		-> (match (simplify x,simplify y) with
				  (Empty,y1) -> y1
				| (x1,Empty) -> x1
				| (x1,y1)    -> Union(x1,y1))
| Product(x,y)		-> (match (simplify x,simplify y) with
				  (Empty,y1)   -> Empty
				| (x1,Empty)   -> Empty
				| (Epsilon,y1) -> y1
				| (x1,Epsilon) -> x1
				| (x1,y1)      -> Product(x1,y1))
| Sequence(x,c)		-> (match simplify x with
				  Empty 	-> Empty
				| Epsilon	-> Epsilon
				| x1 		-> Sequence(x1,c))
| Cycle(x,c)		-> (match simplify x with
				  Empty 	-> Empty
				| Epsilon	-> Epsilon
				| x1 		-> Cycle(x1,c))
| Set(x,c)		-> (match simplify x with
				  Empty 	-> Empty
				| Epsilon	-> Epsilon
				| x1 		-> Set(x1,c))
| MultiSet(x,c)		-> (match simplify x with
				  Empty 	-> Empty
				| Epsilon	-> Epsilon
				| x1 		-> MultiSet(x1,c))
| Atom(0,1,_)		-> Epsilon
| x			-> x;;

(******************* eq_type **********************)
let rec eq_type t1 t2 = if t1=t2 then true else match (t1,t2) with
  (Ident(s,_),Ident(t,_))	-> if atomp s then
				    if atomp t then false
				    else eq_type t1 (get_def t !types)
				   else eq_type (get_def s !types) t2
| (Ident(s,_),y)     		-> eq_type t2 t1
| (Product(x,y),Product(z,t))  	-> eq_type x z & eq_type y t
| (Product(x,y),Ident(s,_))	-> if atomp s then 
					(eq_type x t2 & eq_type y Epsilon)
				   else eq_type t1 (get_def s !types)
| (Product(x,y),Sequence(z,c))	-> if x=z then eq_type y (Sequence(z,sub1 c))
				   else false
					(* update 14/10/92 : previously
					    eq_type x t2 & eq_type y Epsilon *)
| (Product(x,y),z)		-> eq_type x z & eq_type y Epsilon
| (z,Product(x,y))		-> eq_type x z & eq_type y Epsilon
| (Sequence(x,cx),Sequence(y,cy))-> eq_type x y & eq_card cx cy
| (Sequence(x,c),Epsilon)	-> simpl_card c = Eq 0
| (Set(x,c),Epsilon)		-> simpl_card c = Eq 0
| (Epsilon,Sequence(x,c))	-> simpl_card c = Eq 0
| (Epsilon,Atom(k,m,_))		-> if k=0 & m=1 then true else false
| (Set(x,cx),Set(y,cy))		-> eq_type x y & eq_card cx cy
| (MultiSet(x,cx),MultiSet(y,cy))-> eq_type x y & eq_card cx cy
| (Union(a,b),Union(c,d))	-> (eq_type a c & eq_type b d) or
				   (eq_type a d & eq_type b c)
| (Union(_),_)			-> false
| (_,Union(_))			-> false
| (Epsilon,Ident(s,_))		-> false (* update 11/10/88 *)
| (x,Ident(s,_))		-> if atomp s then false
				   else eq_type x (get_def s !types)
| e		  		-> pm "eq_type" e;;

(**********************  filter *************************)
let rec filter env ptype pat = match (ptype,pat) with
  
  (x,Pat_all)              -> []
| (Empty,_)		   -> failwith "filter"
| (Epsilon,Pat_epsilon)	   -> []
| (Epsilon,Pat_prod(c,d))  -> (filter env Epsilon c) @ (filter env Epsilon d)
| (Atom(_,_,_),_)	   -> failwith "filter"
| (LAtom(_,_,_),_)	   -> failwith "filter"
| (Ident(s,po),p)       -> (try let sd = if boundp s !types then
						get_def s !types 
					     else get_def s env in
				match p with
				 Pat_id s2 -> if s2=s then if po=No then []
					 else err "filter" ("change name "^s^" in pattern") ()
	                         	else if boundp s2 (!types@env) then
 (* s2 is a defined type or variable *)	 filter env sd p
 (* s2 is a new variable *)		 else [(s2,eval env (Ident(s,po)))]
				 | _      -> filter env sd p
	                       with Failure "find" ->
				err "filter" "unknown ident" s)
| (Union(x,y),p)	-> (match p with
				Pat_id s -> if boundp s env then
				 user_err "pattern variable already defined" s;
					    if boundp s !types then
                                 		(try filter env x p with
						    Failure "filter" ->
						     filter env y p)
					   else [(s,eval env ptype)]
			      | _       -> try filter env x p with
					   Failure "filter" -> filter env y p)
| (x,Pat_id s)           -> if boundp s !types then
                                 if eq_type x (get_def2 s !types)
				 then [] else (match x with
				  Sequence(Ident(y,_),c) -> if y=s then []
					else failwith "filter"
				| _ -> failwith "filter")
			       else if boundp s env then 
				 err "filter" "pattern variable is already defined" s
			       else [(s,eval env x)]
| (Product(a,b),
   Pat_prod(c,d))	-> (filter env a c) @ (filter env b d)
| (OProduct(a,b),y)	-> filter env (Product(a,b)) y
| (Set _,Pat_prod _)	-> failwith "filter"
| (MultiSet _,Pat_prod _)-> failwith "filter"
| (Sequence(l1,c),y) 	-> if lowest c = 0 then try filter env Epsilon y with
			    Failure "filter" -> filter env 
				(Product(l1,Sequence(l1,sub1 c))) y
				   (* Ge(c,1) -> sub1 c : update 11/10/88 *)
			   else filter env (Product(l1,Sequence(l1,sub1 c))) y
| (_,Pat_epsilon)	-> failwith "filter"  (* update 18/3/89 *)
| (Set(_,c),Pat_emptyset) -> if lowest c = 0 then [] else failwith "filter"
| (MultiSet(_,c),Pat_emptyset) ->if lowest c = 0 then [] else failwith "filter"
| x                     -> pm "filter" x ;;

(************************ subtract ***************************)
let rec subtract t0 t1 = 		(* t0 : initial type *)
					(* t1 : matching type *)
(* returns (t0-t1,z) with z=true then t0-t1 <> t0 *)
match t0 with

  Union(a,b)	-> (match t1 with
	  
	  Union(c,d) -> let (t2,z) = subtract t0 c in
			 let (t3,zz) = subtract t2 d in
			  (t3,z or zz)
	| _	-> let (a1,z) = subtract a t1 in
		    if z then 
		     if a1 = Empty then (b,true)
		     else (Union(a1,b),true)
		    else let (b1,zz) = subtract b t1 in 
		     if b1 = Empty then (a,zz)
		     else (Union(a,b1),zz))

| Product(a,b)	-> (match t1 with

	  Product(c,d)	-> let (a1,z) = subtract a c in
			    if z = false then (t0,false)
			    else let (b1,zz) = subtract b d in
			     if zz = false then (t0,false)
			     else (match (a1,b1) with
				  (Empty,Empty)	-> (Empty,true)
				| (Empty,_)	-> (Product(a,b1),true)
				| (_,Empty)	-> (Product(a1,b),true)
				| _		-> (Union(Product(a1,b),
						    Product(c,b1)),true))
	| OProduct(c,d)	-> subtract t0 (Product(c,d)) (* PZ 04/05/92 *)
	| Empty		-> (t0,false)
	| Ident(s,po)	-> if mem_assoc s !types then
			    if atomp s
			     then subtract t0 (Product(Ident(s,po),Epsilon))
			    else subtract t0 (get_def s !types)
			   else (t0,false)
	| Epsilon	-> (t0,false) (* update 18/3/89 *)
	| _		-> subtract t0 (Product(t1,Epsilon)))
			   (* for Sequence, Cycle, Set, MultiSet *)

| OProduct(a,b)	-> (match subtract (Product(a,b)) t1 with
	  (Product(c,d),zz) -> (OProduct(c,d),zz)
	| xx		    -> xx)

| Sequence(a,c)	-> (match t1 with
	
	  Sequence(b,d)	-> if eq_type a b then let c1 = sub_card c d in
			    if eq_card c c1 then (t0,false)
			    else if c1 = None then (Empty,true)
			    else (Sequence(a,c1),true)
			   else (t0,false)
	| _		-> if lowest c = 0 then subtract 
			    (Union(Epsilon,Product(a,Sequence(a,sub1 c)))) t1
			   else subtract (Product(a,Sequence(a,sub1 c))) t1)

| Set(a,c)	-> (match t1 with

	  Set(b,d)	-> if eq_type a b then let c1 = sub_card c d in
			    if eq_card c c1 then (t0,false) 
			    else if c1 = None then (Empty,true)
			    else (Set(a,c1),true)
			   else (t0,false)
	| Emptyset	-> if lowest c = 0 then (Set(a,sub_card c (Eq 0)),true)
			   else (t0,false)
	| Ident(s,_)	-> if atomp s then (t0,false) 
			   else subtract t0 (get_def s !types)
	| _		-> (t0,false))

| MultiSet(a,c)	-> (match t1 with

	  MultiSet(b,d)	-> if eq_type a b then let c1 = sub_card c d in
			    if eq_card c c1 then (t0,false) 
			    else if c1 = None then (Empty,true)
			    else (MultiSet(a,c1),true)
			   else (t0,false)
	| Ident(s,_)	-> if atomp s then (t0,false) 
			   else subtract t0 (get_def s !types)
	| _		-> (t0,false))

| Cycle(a,c)	-> (match t1 with

	  Cycle(b,d)	-> if eq_type a b then let c1 = sub_card c d in
			    if eq_card c c1 then (t0,false) 
			    else if c1 = None then (Empty,true)
			    else (Cycle(a,c1),true)
			   else (t0,false)
	| Ident(s,_)	-> if atomp s then (t0,false) 
			   else subtract t0 (get_def s !types)
	| _		-> (t0,false))

| Empty		-> (t0,false)

| Epsilon	-> (match t1 with
	
	  Epsilon	-> (Empty,true)
	| Product(_,_)	-> subtract (Product(Epsilon,Epsilon)) t1
	| Sequence(a,c)	-> (match simpl_card c with
				  Eq 0	-> (Empty,true)
				| None	-> (t0,false)
				| _	-> err "subtract" "t1 > t0" (t0,t1))
	| Ident(s,_)	-> if atomp s then (t0,false) 
			   else subtract t0 (get_def s !types)
	| _		-> nyi "subtract" "Epsilon - other")

| Ident(s,_)	-> if t1 = t0 then (Empty,true)
		   else if atomp s then (match t1 with
			  Ident(t,No)  -> if s=t then (Empty,true)
					  else if atomp t then (t0,false)
					  else subtract t0 (get_def t !types)
			| Ident(t,_)   -> if atomp t then (t0,false)
				          else subtract t0 (get_def t !types)
			| Product(c,d) -> subtract (Product(t0,Epsilon)) t1
			| Epsilon      -> if atom_size s = 0 then (Empty,true)
					  else (t0,false)
			| Union(c,d)   -> let (sc,zc) = subtract t0 c
                                          in if zc then (sc,zc)
                                             else subtract t0 d
			| _	  -> nyi "subtract" "atom - other")
		   else subtract (get_def s !types) t1
| Mark(_)	-> (t0,false)
| _		-> pm "subtract" t0;;

(************************ struct_of_pat ************************)
let rec struct_of_pat = function
  Pat_id s        -> Ident(s,No)
| Pat_epsilon	  -> Epsilon
| Pat_emptyset	  -> Emptyset
| Pat_prod(a,b)	  -> Product(struct_of_pat a,struct_of_pat b)
| x		  -> err "struct_of_pat" "pattern not valid" x ;;

(*************************** struct_of_expr **************************)
let rec struct_of_expr = function
  Exp_id s	    -> Ident(s,No)
| Exp_constr(c,l)   -> transform(c,map struct_of_expr l)
| x		    -> nyi "struct_of_expr" "other than ident or constructor";;

(*************************** vars_of_expr ************************)
let rec vars_of_expr = function
  Exp_id s		-> [s]
| Exp_constr(c,l)	-> flat (map vars_of_expr l)
| Call(f,l)		-> flat (map vars_of_expr l);;

(*************************** vars_of_struct **********************)
let vars_of_struct = foo where rec foo = function
  Union(a,b)		-> (foo a) @ (foo b)
| Product(a,b)		-> (foo a) @ (foo b)
| Sequence(a,_)		-> foo a
| Cycle(a,_)		-> foo a
| Set(a,_)		-> foo a
| MultiSet(a,_)		-> foo a
| Ident(s,_)		-> [s]
| _			-> [];;

let rec vars_of_pattern = function
  Pat_id(s)	-> if boundp s !types then [] else [s]
| Pat_prod(a,b)	-> (vars_of_pattern a)@(vars_of_pattern b)
| _		-> []
;;

(*************************** vars_of_cond ************************)
let vars_of_cond = function
  Equal(s,p)		-> vars_of_struct s
| Predicate(s,t)	-> [s;t]
;;
(*************************** vars_of_instr ************************)
let rec vars_of_instr = function
  Cases(l,m)		-> l @ (flat (map vars_of_instr (map snd m)))
| Casetype(s,m)		-> s :: (flat (map vars_of_instr (map snd m)))
| Instructions(l)	-> flat (map vars_of_instr l)
| IfThenElse(c,i,j)	-> (vars_of_cond c) @ (vars_of_instr i) @ (vars_of_instr j)
| Forall(s,t,i)		-> [s;t] @ (vars_of_instr i)
| Forone(s,t,i)		-> [s;t] @ (vars_of_instr i)
| Expression(e)		-> (match e with
				  Call(f,l) -> flat (map vars_of_expr l)
				| _         -> []) (* same cost here *)
| Repeat(i,_)		-> vars_of_instr i
| Empty_instr		-> []
;;
(*************************** replace **************************)
let replace ids sts = foo where rec foo = function
  []	-> []
| x::y	-> if mem x ids then let n = index x ids in 
		(vars_of_struct (nth sts n))@(foo y)
	   else x::(foo y);;

(************************** has_double ***************************)
let rec has_double = function
  []	    -> false
| x :: l    -> if mem x l then true else has_double l;;

(************************** body *********************************)
(* returns the name of a function equivalent to f, defined in !functions *)
let rec repr_of f = 
	if mem_assoc f !functions then f (* this must be first not to loop
					    if ([f],f) is in !fnames *)
	else if mem_assoc [f] !fnames 
	then repr_of (assoc [f] !fnames)
	else user_err ("Undefined function: "^f) ""
;;

let rec body f = snd (snd (get_def (repr_of f) !functions)) 
;;

(************************** params *******************************)
let params f = fst (get_def (repr_of f) !functions)
;;

let input_type_aux = map (fun (_,Ident(y,_)) -> y | _ -> "");;

let input_type f = input_type_aux (params f);;

let def_of_comp f = try (rev_assoc f !fnames) 
	with Failure _ -> user_err ("Undefined function: "^f) "";;

let rec input_type1 f = if mem_assoc f !functions then
  (match input_type f with
    [s] -> s
  | l   -> err "input_type1" "number of arguments <> 1" (list_length l))
else input_type1(hd(def_of_comp f));;

let output_type f = fst (snd (get_def (repr_of f) !functions));;

(************************** join *********************************)
let join env (fp,_) rp = (fp,eval env (struct_of_expr rp));;
				    (* to join formal and real parameters *)

(**************************** mu ***************************)
let mu f = if boundp f !measures 
then let m = get_def f !measures in match m with
    NumCost(n) -> Gf_num(n)
  | ParCost(s) -> Gf_par(s)
else (if not boundp f !functions then
	if not mem f !id_not_def then (
	   id_not_def := f :: !id_not_def;
	   war ("function ``"^f^"'' not defined, cost assumed to be 0."));
	Gf_num(0));;

let filter_union id = foo false where rec foo found = function
  Union(x,y)	-> (match x with
			  Ident(k,_)	-> if k=id then (match y with
				  Union(_,_) -> foo true y
				| _	     -> (Union(foo true y,Empty)))
					   else (Union(x,foo found y))
			| _		-> (Union(x,foo found y)))
| y		-> if found then y
		   else err "filter_union" "type not found" id
;;

