##
##    Title: 	luo/solve version 2.0
##    Created:	Wed Oct 26 11:54:19 1988
##    Author: 	Paul Zimmermann
##		<zimmerma@bandol>
##
##
##    Modified: 	Tue Feb 14 15:20:59 1989
##    Author: 	Paul Zimmermann
##
##    Modified: 	Tue May 23 12:07:23 1989
##    Author: 	Bruno Salvy and Paul Zimmermann
##    Modification: installation of the new version of equivalent
##
##
##    Modified: 	Thu Feb 23 10:26:46 1995
##    Author: 	Bruno Salvy
##    Modification: adapted to V.3
##


# Luo printlevel
luo_level:=1:
set_luo_level:=proc(n) 
global luo_level,infolevel; 
   luo_level:=n; infolevel[luo]:=n;
end:

`luo/tocite` := {}:

`luo/cite` := proc(s) 
global `luo/tocite`;
    `luo/tocite`:=`luo/tocite` union {s}; s 
end:

`luo/citations` := proc() 
global `luo/tocite`;
    `luo/tocite`; 
    `luo/tocite`:={}; 
    if ""={} then ` ` else op("") fi
end:

macro(`equivalent/equiv`=readlib(`equivalent/equiv`),
`dev/add`=readlib(`dev/add`),
`dev/multbyreal`=readlib(`dev/multbyreal`),
`dev/print`=readlib(`dev/print`),
`dev/divide`=readlib(`dev/divide`),
`dev/dev`=readlib(`dev/dev`),
`dev/instanc`=readlib(`dev/instanc`),
`dev/prd`=readlib(`dev/prd`),
`dev/length`=readlib(`dev/length`),
isolate=readlib(isolate));

# this function is called from Caml
# type_eqs: equations for objects
# compl_eqs: equations for complexity descriptors
# toana: what is to be analyzed
# sys0: initial equations
luo1 := proc(type_eqs,compl_eqs,toana,sys0)
local sol;
global luo_sol0;
    `luo/verify`(type_eqs union compl_eqs);
    luo_sol0:=map(proc(eq) `luo/transform`(op(1,eq))=op(2,eq) end,solve(sys0));
    if luo_sol0=NULL then ERROR(`No solution for values in z=0`) fi;
    if nops([luo_sol0])>=2 then 
	ERROR(`Too many solutions for values in z=0`) 
    fi;
    `luo/indic`(`Solving counting generating functions`);
    if luo_level>=3 then lprint(`   Values in z=0 are:`); print(luo_sol0) fi;
    sol:=luo(type_eqs);
    if compl_eqs<>{} then `luo/indic`(`Solving complexity descriptors`) fi;
    sol:=sol union luo(eval(subs(sol,compl_eqs)));
    `luo/assign`(sol);
    `luo/indic`(`Solution is`);
    if luo_level>0 then map(proc(f) print(``.f.`(z)`=f(z)) end,toana) fi
end:

# translates x0 into x(0)
`luo/transform`:=proc (x)
    substring(x,1..length(x)-1)(0)
end: # `luo/transform`

`type/int` := proc(x)
    type(x,function) and op(0,x)=int
end:

# solve one equation
mysolve:=proc(id,rm)		# id: A(z) ou A(z,u)     rm: right member
local res,v0;
    if has(rm,{int,intg}) then
	if luo_level>=5 then 
    lprint(`     This is an integral equation. Trying to differentiate it:`) 
	fi;
	v0:=`luo/value_in_zero`(op(0,id));
	if v0=NULL then v0:=`luo/cost_in_zero`(op(0,id),rm) fi;
	diff(id,z)=diff(rm,z);
	if luo_level>=5 then print(") fi;
	res := traperror([dsolve({",op(0,id)(0)=v0},id)]);
	if "=lasterror or "=[] then
	    if luo_level>=5 then 
    lprint(`     Derivation failed. Trying to isolate the integrals ...`) 
	    fi;
	    # on essaie quand me^me qd il y a une seule inte'grale
	    indets(rm,int);
	    if nops(")>1 then
		ERROR(`I can not solve this integral equation: `,id=rm)
	    else
		isolate(id=rm,op("));
		if luo_level>=5 then print(") fi;
		if luo_level>=5 then lprint(`     After derivation:`) fi;
		map(diff,",z);
		if luo_level>=5 then print(") fi;
		res := traperror([dsolve({",op(0,id)(0)=v0},id)]);
		if "=lasterror or "=[] then
		    ERROR(`I can not solve this integral equation: `,id=rm)
		fi
	    fi
	fi;
	`luo/polish`(op(res),op(0,id),rm)
    else
	traperror(degree(subsQL(rm),id));
	if "=3 then RETURN({id=RootOf(_Z-subs(id=_Z,rm))}) fi;
	#    res:=traperror(solve({id=rm},{op(0,id)}));
	res:=traperror(solve({id=rm},{id}));
	if res=lasterror then ERROR(`I can not solve this equation: `,id=rm)
	elif res=NULL and has(rm,{Q,L}) then solve({id=subsQL(rm)},{id})
	else res
	fi
    fi;
end:



# isolates v(z) and finds constants
`luo/polish`:=proc (x,v,rm)
local sol,cval,v0;
    sol:=isolate(x,v(z));
    if luo_level>=5 then lprint(`     General solution is `,sol) fi;
    if has(sol,{_C1}) then
	v0:=`luo/value_in_zero`(v);
	if v0=NULL then v0:=`luo/cost_in_zero`(v,rm) fi;
	if luo_level>=5 then lprint(`     Value of `.v.`(0) is `.v0) fi;
	subs(z=0,v(0)=v0,x);
	cval := solve(",_C1);
	if cval=NULL then ERROR(`No constant found in`,x,`to give`,v0) fi;
	if luo_level>=5 then
	    lprint(`     Constant _C1 is taken to be `,cval) fi;
	sol:=simplify(subs(_C1=cval,sol))
    fi;
    {sol}
end: # `luo/polish`

# determine initial values
`luo/value_in_zero`:=proc (id)
    op(map(proc(x,y) if op(1,x)=y then op(2,x) fi end,luo_sol0,id(0)));
end: # `luo/value_in_zero`

# idem for complexity descriptors
`luo/cost_in_zero`:=proc(id,t)	# for a complexity descriptor
      	       	     	      	# t is the right member of the equation
    subs(id(z)=0,t);
    taylor(",z);
    convert(",polynom);
    coeff(",z,0);
end:

`type/posnum` := proc(x) 
    if type(x,integer) or type(x,fraction) then evalb(x>=0) 
    else false fi 
end:

`series/MP` := proc(f,z)
local s,i;
    s:=0;
    for i to Order-1 do
	s := s + subs(z=z^i,f)/i
    od;
    series(exp(s),z);
end:

`series/PPolya` := proc(f,z)
local s,i;
    s:=0;
    for i to Order-1 do
	s := s + (-1)^(i+1)*subs(z=z^i,f)/i
    od;
    series(exp(s),z);
end:

valid_aux := proc(y)
    series(y,z);
    simplify(");	# for example file paren.adl
    convert(",polynom);
    type(",polynom(polynom(posnum),z))
end:

# eliminate bad solutions
valid := proc(x)	# returns false if x is not valid
      	       	     	# x := <gf> or <id>=<gf> or {<id>=<gf>,...,<id>=<gf>}
local y;
    if type(x,set) then
	true;
	for y in x while " do
	    valid(y)
	od
    elif type(x,`=`) then valid(op(2,x))
    else
	y := subsQL(x);
	y := subs(map(proc(x) x=1 end,marks),y);
	traperror(valid_aux(y));
	if "=false then 
	    if luo_level>=3 then lprint(`   Eliminating solution: `,x) fi;
	    RETURN(false)
	fi;
# bug in glimit : -1/2*(-RootOf(-2*z*_Z**2+_Z-z)+z)/z/RootOf(-2*z*_Z**2+_Z-z)
#  if has(glimit(y,z=0),{`undefined`,infinity}) then
#      if luo_level>=3 then lprint(`   Eliminating solution: `,x) fi;
#      RETURN(false)
#  fi;
	RETURN(true)
    fi
end:

treq := proc(eq)		# id(z) = rhs -> id = proc(z) rhs end
    op(1,eq);
    op(0,") = subs(r=op(2,eq),proc(z) r end);
end:

check_functional := proc(sys1)	# when there is some toto(z^k) --> ERROR
    indets(sys1,function);
    map(proc(x) op(1,x); if not member(op(0,x),{Q,L}) and type(",`^`) and 
      	       	  op(1,")=z then x fi end,");
    if "<>{} then
	lprint(`Solving functional equations is not yet implemented:`);
	ERROR(`The system was: `,sys1);
    fi
end:

# general solver (system of ordinary or differential equations)
luo:=proc (sys)	      	      #  sys : { eq1, ... , eqN }
      	       	     	      #  eq : id(z) = rm
local sys1,sol,li,liz,found_one,eq,n,i,id,rhs,s,eq1;
    if luo_level>=3 then
	lprint(`   The system to solve is:`);
	print(sys)
    fi;
    sys1:=sys; sol:={};
  # determines explicit and separate equations
    do
	if luo_level>=5 then
	    lprint(`     It remains to solve:`);
	    print(sys1)
	fi;
	n:=nops(sys1);
	liz:=map(<op(1,x)>,sys1);		# {a(z),b(z),...}
	li:=map(<op(0,x)>,liz);		# {a,b,...}
	found_one:=false;
	for i to n while found_one=false do
	    eq:=op(i,sys1); id:=op(0,op(1,eq)); rhs:=op(2,eq);
	    if not has(rhs,li minus {id}) then
		if not has(rhs,id) then
		    # explicit equation
		    found_one:=true;
		    if luo_level>=4 then
			lprint(`    Explicit equation: `,eq) fi;
		    eq1:=treq(eq);
		    sol:=sol union {eq1};
		    sys1:=eval(subs(eq1,subsop(i=NULL,sys1)));
		else # right member contains only variable in left member
		    if luo_level>=4 then
			lprint(`    Separate equation: `,eq) fi;
		    indets(eq,function);
		    map(subs(Id=id,proc(x) if op(0,x)=Id then op(x) fi end),");
		    if hastype(",`^`) then
			# eq contains at least one id(z^k) with k<>1
			lprint(cat(id,
		`(z) is the solution of the following functional equation:`));
			print(eq);
			s:={};	  	  
		    else
			s:=mysolve(op(1,eq),rhs);
			if nops([s])>1 then 
			    s:=op(map(proc(x,sol) if valid(subs(sol,
				op(2,op(x)))) then x fi end,[s],sol));
			fi;
			if s<>lasterror and nops([s])=1 then
			    found_one:=true;
			    if luo_level>=4 then 
				lprint(`    Solution of `,
				    id(z)=rhs,` is `,op(s)) fi;
			    sol:=sol union map(treq,s);
			    sys1:=eval(subs(op(map(treq,s)),
				subsop(i=NULL,sys1)));
			elif s<>lasterror and nops([s])>1 then
			    ERROR(`too many solutions found for`,eq)
			else
			    ERROR(`no valid solution found for`,eq)
			fi;
		    fi;
		fi;
	    fi;
	od;
	if found_one=false then break fi;
    od;
    # there are no more explicit or separate equations
    if luo_level>=4 then 
	lprint(`    There are no more explicit or separate equations.`);
	lprint(`    It remains to solve:`);
	print(sys1)
    fi;
    if sys1={} then sol
    else
	check_functional(sys1);
	if has(sys1,int) then s:=`luo/dsolve`(sys1,liz)
	else s:=solve(sys1,liz)
	fi;
	if s=NULL then ERROR(`No solution found to `,sys1) fi;
	# check NULL before because has(NULL,...) gives an error
	if has(s,RootOf) then s:=readlib(allvalues)(s,'d') fi;
	s:=op(map(proc(x) if valid(x) then x fi end,[s]));
	if nops([s])>1 then ERROR(`Too many solutions to `,sys1,[s])
	else 
	    if luo_level>=4 then
		lprint(`    Solution of implicit equations is:`);
		print(s)
	    fi;
	    sol union map(treq,s)
	fi;
    fi;
end: # luo

`luo/dsolve` := proc(sys,vars)
local eq,v,eq1,sys1,vars1;
    traperror(dsolve(sys,vars));
    if "<>lasterror and "<>FAIL then "
    else
	for eq in sys do
	    v := op(1,eq);
	    traperror(solve(eq,v));
	    if "<>lasterror and "<>FAIL and nops(["])=1 then
		eq1 := v=";
		sys1 := subs(eq1,sys minus {eq});
		vars1 := vars minus {v};
		print(`It remains to solve `,sys1,vars1);
	    else
		lprint(`Maple is unable to solve the following system:`,sys)
	    fi
	od
    fi
end:

# tries to simplify x
mysimplify:=proc(x)
    simplify(subsQL(x));
    if has(",I) then x else " fi
end:

# assign and remember assigned names
`luo/assign`:=proc (sys)
local eq,v;
global luo_names;
    for eq in sys do
	v:=op(1,eq);      
	if type(v,function) and op(0,v)=evaln then v:=op(1,v) fi;
	if not type(v,string) then 
	    ERROR(`Left member of equation is not a string: `,v)
	fi;
	if type(v,procedure) then
	    ERROR(`The symbol "`.v.`" is the name of a Maple procedure`)
	fi;
	if not type(luo_names,set) then luo_names:={} fi;
	luo_names:=luo_names union {`#`.`v`};
	assign(v=eval(op(2,eq)));
    od;
end: # `luo/assign`

`luo/maple_keywords` := 
    {list,set,string,seq,`minus`,constants} minus {true,false}:

is_a_maple_keyword := proc(x)
    member(x,`luo/maple_keywords` union {Q,L,z,u}) or assigned(x)
end:

# verify that no keywords are used
`luo/verify`:=proc (x)
    map(proc(k,e) if has(e,k) then k fi end,`luo/maple_keywords`,x);
    if "<>{} then ERROR(`Maple keywords used`,") fi;
    map(proc(e) op(0,op(1,e)) end,x) intersect {Q,L,z,u};
    if "<>{} then 
	ERROR(`You are not allowed to use these special variables`,op("))
    fi;
end: # `luo/verify`

unassign := proc(t)
    t:='evaln(t)';
end:

# called between two analyses
`luo/unassign` := proc()
global luo_names;
    if type(luo_names,set) then
	map(proc(t) unassign(substring(t,2..500)) end,luo_names)
    fi;
    luo_names := {};
end:

Equiv:=proc(fct,x)
local i, p, split; #it would be cleaner to use N as a local 
	#variable here, but it would lose some remember tables
    if nargs=3 then p:=args[3]
    else p:=1
    fi;
    if nops(indets(fct,name) minus {constants})=1 then
	`equivalent/equiv`(fct,p,x)
    else
	split:=separvar(fct,x); # for some strange reason, the variable split
	0;                    # was necessary, though for i in separvar(fct,x)
	for i in split do       # should have been enough.
	    `equivalent/equiv`(i[2],p,x);
	    if not has(",_saddlepoint) then
		`dev/add`("",`dev/multbyreal`(",i[1]))
	    else
		if nops(split)>1 then
		    ERROR(`Cannot handle parameters when saddlepoints occur`)
		else
		    [_saddlepoint,op(2,"),`dev/multbyreal`(op(3,"),i[1])]
		fi
	    fi
	od
    fi
end:

separvar:=proc(expr,x)
local ind,i;
    ind:=indets(expr,name) minus {constants};
    if nops(ind)=0 then RETURN([[expr,1]])
    elif nops(ind)=1 then
	if ind={x} then RETURN([[1,expr]])
	else RETURN([[expr,1]])
	fi
    elif type(expr,function) then
	expand(expr);
	if "<>expr then RETURN(separvar("))
	else ERROR(`Les variables ne se separent pas bien`)
	fi
    elif type(expr,`^`) then
	expand(expr);
	if "<>expr then RETURN(separvar(",x)) fi;
   	ind:=map(proc(x) if type(x,name) then x fi end,indets(op(2,expr)));
	if nops(ind)=0 then
	    RETURN(map(proc(x,y) [x[1]^y,x[2]^y] end,
		separvar(op(1,expr),x),op(2,expr)))
	elif nops(ind)=1 then
	    separvar(op(1,expr),x);
	    if nops(")=1 and member(1,op("),'i') then
		if i=1 and ind={x} then RETURN([[1,"[1][2]^op(2,expr)]])
		elif i=2 and ind<>{x} then RETURN([["[1][1]^op(2,expr),1]])
		fi
	    fi
	fi;
	ERROR(`Les variables ne se separent pas bien`)	
    elif type(expr,`*`) then
	[[1,1]];
	for i to nops(expr) do
	    separvar(op(i,expr),x);
	    proc(x,y) ['op([[x['i'][1]*y['j'][1],x['i'][2]*y['j'][2]]
				$'i'=1..nops(x)])'$'j'=1..nops(y)] end(","");
	od;
	RETURN(")
    elif type(expr,`+`) then
	map(proc(x,y) op(separvar(x,y)) end,[op(expr)],x);
	map(proc(x) local i;member(1,x,'i');if " then[",i] else " fi end,");
	if (not has(",false)) and
	    nops(convert(map(proc(x) op(2,x) end,"),set))=1 then
	    if has(",1) then
		RETURN([[1,convert(map(proc(x) x[2] end,""),`+`)]])
	    else RETURN([[convert(map(proc(x) x[1] end,""),`+`),1]])
	    fi
	else RETURN("")
	fi;
    else ERROR(`unknown function`,expr)
    fi
end:

# {k1 + p l1} union {k2 + q l2} \subset {min(k1,k2) + r igcd(l1,l2,k2-k1)}
int_sum := proc()
local k,l,k1,l1,p;
    k:=infinity; l:=0;
    for p in [args] do
	k1:=p[1]; l1:=p[2];
	if k=infinity then k:=k1; l:=l1;
	elif k1<>infinity then
	    l:=igcd(l,l1,k1-k);# this line must precede the following (23/2/89)
	    k:=min(k,k1);
	fi;
    od;
    [k,l];
end:

# {k1 + p l1} plus {k2 + q l2} \subset {k1 + k2 + r igcd(l1,l2)}
int_prod := proc()
local k,l,k1,l1,p;
    k:=0; l:=0;
    for p in [args] do
	k1:=p[1]; l1:=p[2];
	if k<>infinity then
	    if k1=infinity then k:=infinity
	    else
		k:=k+k1;
		l:=igcd(l,l1);
	    fi;
	fi;
    od;
    [k,l];
end:

maxvar0 := proc(e,var)
local k,l;
option remember;
    maxvar(subsQL(e),var);
    k:="[1]; l:=""[2];
    if l<>0 then
	k:=irem(k,l);
	if k<0 then k:=k+l fi;
    fi;
    RETURN(k,l);
end:

# returns [k,l] such that only [z^(k+il)] e <> 0  (i>=0)
# l=0 means that only [z^k] e <> 0
maxvar := proc(e,var) 
    if not has(e,var) then [0,0]
    elif e=var then [1,0]
    elif type(e,{integer,fraction}) then [0,0]
    elif type(e,`+`) then int_sum(op(map(maxvar,[op(e)],var)))
    elif type(e,`*`) then int_prod(op(map(maxvar,[op(e)],var)))
    elif type(e,`^`) then
	if type(op(2,e),integer) then
	    if op(1,e) = var then [op(2,e),0]
	    else 
		maxvar(op(1,e),var);
		[op(2,e)*"[1],"[2]];
	    fi
	elif type(op(2,e),fraction) then maxvar(op(1,e),var)
	else [0,1]
	fi;  
    elif type(e,function)
	then int_sum(op(map(`maxvar/pow`,map(maxvar,[op(e)],var))))
    else ERROR(`Case not treated:`,e)
    fi;
end:

`maxvar/pow` := proc(x)
    if x[2]=0 then [0,x[1]] else x fi;
end:

# translate z^(km) into z^k
subsvar := proc(e,var,m)
    if m=1 then e
    elif type(e,constant) then e
    elif type(e,name) then
	if e=var then var^(1/m) else e fi
    elif type(e,`^`) and op(1,e)=var and type(op(2,e),integer) then
	var^(op(2,e)/m)
    else map(subsvar,e,var,m)
    fi;
end:

subsEquiv := proc(f,var,ord,m,k)
    subs(map(proc(x) x=1 end,marks),f)/var^k;
    [m,Equiv(subsvar(",var,m),var,ord),k]
end:

`dev/printall`:=proc(dev,p)
    if not has(dev,_saddlepoint) then `dev/print`(dev,n,p)
    else
	if op(2,dev)<>infinity then
	    if nops(dev)=3 then
		eval(subs(x=1/(1-_saddlepoint/op(2,dev)),
		    O=proc(x) O(x/_saddlepoint^n) end,
		    `dev/print`(`dev/multbyreal`(op(3,dev),
			1/_saddlepoint^n),x,p)))
	    elif op(3,dev)=1 then
		eval(subs(x=1/(1-_saddlepoint/op(2,dev)),O=proc(x)
		O(x/_saddlepoint^n) end,`dev/print`(`dev/multbyreal`(op(4,dev),
		1/_saddlepoint^n),x,p)))
	    else
		eval(Sum(subs(x=1/(1-_saddlepoint['i']/op(2,dev)),O=proc(x)
		    O(x/_saddlepoint['i']^n) end,`dev/print`(
			`dev/multbyreal`(op(4,dev),1/_saddlepoint['i']^n)
			,x,p)),'i'=1..op(3,dev)))
	    fi
	elif nops(dev)=3 then
	    `dev/print`(op(3,dev),_saddlepoint,p)
	elif op(3,dev)=1 then
	    `dev/print`(op(4,dev),_saddlepoint,p)
	else
	    Sum(`dev/print`(op(4,dev),_saddlepoint['i'],p),'i'=1..op(3,dev))
	fi
    fi;
end:

Dev2dev := proc(Dev,p)
    if op(1,Dev)=1 then
	op(2,Dev)
    else
	`dev/subs`(op(2,Dev),op(1,Dev),p,op(3,Dev))
    fi;
end:

my_printdev := proc(Dev,p)
    `dev/printall`(Dev2dev(Dev,p),p);
end:

my_dividev := proc(dev1,dev2,p,avrg)
local res;
	 # the following should not happen in Luo
# what is this op(3,dev) ? PZ 12/06/91	 
#       if op(3,dev1)<>op(3,dev2) then ERROR(`Non compatible degrees`) fi;
    if dev1[1]<>dev2[1] then ERROR(`Non compatible degrees`) fi;
    if not has(dev1[2],_saddlepoint) and not has(dev2[2],_saddlepoint)
    then res:=`dev/divide`(dev1[2],dev2[2],p)
    elif not has(dev1[2],_saddlepoint) or not has(dev2[2],_saddlepoint)
    then ERROR(`This division is too difficult`)
    elif dev1[2][2]<>dev2[2][2] then
	ERROR(`Cannot divide expansions in different points`)
    elif nops(dev1)=4 or nops(dev2)=4 then
	ERROR(`Cannot divide multiple saddle points`)
    else res:=[_saddlepoint,dev1[2][2],`dev/divide`(dev1[2][3],dev2[2][3],p)]
    fi;
    if nops(dev1)<=2 then avrg:=[dev1[1],res]
    else avrg:=[dev1[1],res,dev1[3]]
    fi;
    if dev1[1]<>1 then res:=`dev/subs`(res,dev1[1],p,dev1[3]) fi;
    `dev/printall`(res,p);
end:

# floating point evaluation
luo_evalf := proc(Dev,p)
local dev;
    dev:=Dev2dev(Dev,p);
    `dev/evalf`(dev);
    `dev/printall`(",p)
end:   

# n --> (n-k)/val
`dev/subs`:=proc (dev,val,p,k)
local i;
    if type(dev,constant) then dev
    else
	if has(dev,_saddlepoint) then
	    print(`In the saddlepoint expansion, please substitute n by (n-`,
		k,`)/`,val);
	    dev
	elif not hastype(dev,boolean) then
	    if has(dev,infinity) then
		`dev/dev`(subs(_Xasy=(_Xasy-k)/val,eval(subs(_Xasytab=_equivX,
		    `dev/instanc`(dev)))),p,p)
	    else
		`dev/dev`(subs(_Xasy=(_Xasy-k)/val,eval(subs(_Xasytab=_equivX,
		`dev/instanc`(dev)+O(`dev/instanc`(`dev/findO`(dev)))))),p,p)
	    fi
	else
	    [true];
	    for i from 2 to nops(dev) do
		[op("),[`Listinf/simplify`(k+val*op(1,op(i,dev))),
		    op(2,op(i,dev))]]
	    od
	fi
    fi
end: # `dev/subs`

`dev/findO`:=proc (dev)
local nd;
    nd:=nops(dev);
    if not type(op(nd-1,dev),list) then
	[op(1,dev),op(nd-1,dev),op(nd,dev)]
    else
	[op(1,dev),procname(op(nd-1,dev)),op(nd,dev)]
    fi
end: # `dev/findO`

# Polya(A,k) = [u^k] Polya(u*A)
Polya:=proc(A)
local k,j,u;
    if nargs=2 then
	k:=args[2];
	coeff(taylor(exp(sum(-(-u)^j*subs(z=z^j,A)/j,j=1..k)),u,k+1),u,k)
    else 'Polya(A)'
    fi;
end:

# PPolya(f) = exp(f(z)-f(z^2)/2+f(z^3)/3-...)
PPolya := proc(f)
local i;
    if type(f,polynom(anything,z)) then
	1;
	for i from 0 to degree(f,z) do
	    " * (1+z^i)^coeff(f,z,i)
	od
    else
	'PPolya(f)'
    fi
end:

# MPolya(A,k) = [u^k] MPolya(u*A)
MPolya:=proc(A,k)
local j,u;
    coeff(taylor(exp(sum(u^j*subs(z=z^j,A)/j,j=1..k)),u,k+1),u,k) 
end:

# MP(f) = exp(f(z)+f(z^2)/2+f(z^3)/3+...)
MP:=proc(f)
local i;
    if type(f,polynom(anything,z)) and has(f,z) then
	1;
	for i to degree(f,z) do
	    " * (1-z^i)^(-coeff(f,z,i))
	od
    else
	'MP(f)'
    fi
end:

# prog_MP(f) = f(z) + f(z^2) + f(z^3) + ...
prog_MP:=proc(f)
local m,ff;
    if type(f,polynom(anything,z)) then
	if type(f,`+`) then ff:=convert(f,list) else ff:=[f] fi;
	0;
	for m in ff do
	    " + icontent(m) * z^degree(m,z) * Q(z^degree(m,z))
	od
    else 'prog_MP(f)'
    fi
end:

# trace function
`luo/indic`:=proc(s)
    if luo_level>=2 then
	lprint(`>>> SOLVER: `.s.` ...`);
	lprint();
    fi;
end:

subsQL:=proc(x)
    eval(subs(Q=<1/(1-t)>,L=<log(1/(1-t))>,x));
end:

intg:=proc(expr,var,k)		# k-th integral of expr wrt var
local u;
    if k=0 then expr
    else
	u := gensym();
	subsQL(expr);
	subs(var=u,");
	int(",u=0..var);
	intg(",var,k-1)
    fi
end:

last_number := 0:
gensym := proc()		# generates a new symbol u1,u2,...
global last_number;
    do
	last_number := last_number+1;
	if not has({anames()} union {unames()},u.(last_number)) then break fi
    od;
    u.(last_number)
end:

# compute and print the k-th moment associated to mark m in s
moment := proc (m,taus,s,k,ord,evlf)
local f,g,avrg,mv,res;
    f:=subs(map(proc(x) x=1 end,marks minus {m}),taus(z));
    g:=subs(map(proc(x) x=1 end,marks minus {m}),s(z));
    if luo_level>=3 then 
	lprint(`   Marked generating function for `.taus.` is: `);
	print(f);
    fi;
    mv:=[maxvar0(f,z)];
    res:=`luo/moment`(m,f,g,k,ord,avrg);
    print(res);
    if mv[2]<>1 then 
	lprint(`for n mod `.(mv[2]).` = `.(mv[1]).`, and 0 otherwise.`) 
    fi;
    if evlf then 
	lprint(`Floating point evaluation:`);
	luo_evalf(avrg,ord)
    fi;
    res
end:

# computes the k-th moment associated to mark m in g.f. f
# divides by the number of elements in g
`luo/moment` := proc(m,f,g,k,ord,avrg)
local a,b,an,bn,ff,gg,i,mv;
    mv := [maxvar0(f+g,z)];
    i := mv[2];
    ff := subsvar(f,z,i);
    gg := subsvar(g,z,i);
    subsQL(ff);
    to k do m*diff(",m) od;
    a:=subs(m=1,");
    if luo_level>=3 then
	lprint(`   After differentiation `,k,` times, we obtain:`);
	print(subs(z=z^i,a))
    fi;
    an:=[i,Equiv(a,z,ord)];
    if luo_level>=3 then
	lprint(`   Asymptotic expansion of coefficients:`);      
	print(my_printdev(an,ord))
    fi;
    b:=subs(m=1,subsQL(gg)); # if not subsQL -> Bug M in bugsV1.4
    if luo_level>=3 then
	lprint(`   The generating function without mark is:`);
	print(subs(z=z^i,b)) 
    fi;
    bn:=[i,Equiv(b,z,ord)];		# Equiv(exp(L(z)),z,2) loops
    if luo_level>=3 then
	lprint(`   Asymptotic expansion of coefficients:`);   
	print(my_printdev(bn,ord))
    fi;
    my_dividev(an,bn,ord,avrg);
end:

variance := proc(m,taus,s,ord,evlf)
local f,g,M1,M2,M1bis,var,prd,mv,Ord;
    f:=subs(map(proc(x) x=1 end,marks minus {m}),taus(z));
    g:=subs(map(proc(x) x=1 end,marks minus {m}),s(z));   
    for Ord from ord to ord+4 do
	`luo/moment`(m,f,g,1,Ord,'M1');
	M1:=op(2,M1);
	if luo_level>=2 then
	    lprint(`  Moment of order 1 is:`);
	    print(`dev/printall`(M1,Ord+1));
	fi;
	`luo/moment`(m,f,f,1,Ord,'M1bis');
	M1bis:=op(2,M1bis);
	`luo/moment`(m,f,g,2,Ord,'M2');
	M2:=op(2,M2);
	if luo_level>=2 then
	    lprint(`  Moment of order 2 is:`);
	    print(`dev/printall`(M2,Ord+1));
	fi;
	prd:=`dev/prd`(M1,M1bis);
	if luo_level>=3 then
	    lprint(`   V = M2 - M1^2`);
	    print(`M2 =`,`dev/printall`(M2,Ord+1));
	    print(`M1^2 =`,`dev/printall`(prd,Ord+1));
	fi;
	`dev/multbyreal`(prd,-1);
	var:=`dev/add`(M2,");
	if `dev/length`(var)>ord then break 
	elif luo_level>=2 then
	    lprint(`  Not enough terms. Trying with a larger order.`);
	fi
    od;
    print(`dev/printall`(var,ord));
    mv:=[maxvar0(f+g,z)];
    if mv[2]<>1 then
	lprint(`for n mod `.(mv[2]).` = `.(mv[1]).`, and 0 otherwise.`)
    fi;
    if evlf then 
	lprint(`Floating point evaluation:`);
	luo_evalf([1,var],ord)
    fi;
end:
   
`dev/evalf`:=proc (dev)
local i;
    if has(dev,_saddlepoint) then
	if nops(dev)=3 then
	    [_saddlepoint,op(2,dev),procname(op(3,dev))]
	else
	    [op(1..3,dev),procname(op(4,dev))]
	fi
    elif hastype(dev,boolean) then
	[true];
	for i from 2 to nops(dev) do
	    [op("),[op(1,op(i,dev)),procname(op(2,op(i,dev)))]]
	od
    else
	`dev/evalf/doit`(dev)
    fi;
end: # `dev/evalf`

`dev/evalf/doit`:=proc (dev)
local res, i;
    if not type(dev,list) then evalf(dev)
    else
	res:=[op(1,dev)];
	for i from 2 by 2 to nops(dev) do
	    res:=[op(res),procname(op(i,dev)),op(i+1,dev)]
	od;
	RETURN(res)
    fi
end: # `dev/evalf/doit`

texp := proc (x, k) sum(x^'i'/'i'!,'i' = 0 .. k) end:

Diag := proc(x) subs(z=z^2,x) end:

save `Solve.m`;
quit
