##
##    Modified: 	Thu Jun 20 15:42:35 1991
##    Authors: 	Michele Soria and Paul Zimmermann

macro(infsing=readlib(infsing),
gdev=readlib(gdev),
glimit=readlib(glimit),
evalr=readlib(evalr));

my_infsing := proc(f)
   infsing(f,z,0);
   op(1,op(1,"))
end:

distribution := proc(m,s)
   distrib(subs(map(proc(x) x=1 end,marks minus {m}),s(z)),m)
end:

distrib := proc(Expr,m)   
local expr,f,path,rho,rhof,st,w,C,w0;
   expr := eval(subs(exp=Exp,Expr)); # to simplify exp(L(..)) in Q(..)
   f := op(1,search_marked(expr,m,[]));
   path:= op(2,search_marked(expr,m,[]));
   rho := my_infsing(subs(m=1,expr));
   rhof := my_infsing(f);
   st := critical_status(f,rho,rhof);
   C := op(0,f);			# marked constructor
   w := op(f);
   w0 := traperror(subsQL(subs(z=rho,w)));
   if lasterror=`division by zero` then w0:=infinity fi;
   if st=`critical` then
      if C=Q then
      	 userinfo(2,luo,`Ref: [Soria90,p.125] limit distribution for Q(w) when Q is critical`);
      	 if my_infsing(op(f))<>rho then
      	    print(`The constructor Q is reductor ==> Q(w) has a pole`);
	    gaussian(n);
	 else
      	    userinfo(2,luo,`The constructor Q is non reductor`);
	    if is_algebraic(w) then
	       userinfo(2,luo,`and the singularity of `,w,` is algebraic ==>`);
	       rayleigh();
	    fi
	 fi
      elif C=L then
      	 userinfo(2,luo,`Ref.: [Soria90,p.126] limit distribution for L(w) when L is critical`);
      	 if my_infsing(op(f))<>rho then
      	    print(`The constructor L is reductor ==> L(w) has a logarithmic singularity`);
	    gaussian(n);
	 else
      	    userinfo(2,luo,`The constructor L is non reductor`);
	    if is_algebraic(w) then
	       userinfo(2,luo,`and the singularity of `,w,` is algebraic ==>`);
	       gauss_abs()
	    fi
	 fi
      fi
   elif st=`above critical` then # C is necessary E
      is_logarithmic(w,rho);
      if op(1,") then
      	 RETURN(gaussian(op(2,")*log(n)))
      fi;
      is_polynomial(w);
      if op(1,") then
      	 gaussian(n/op(2,"));
      fi;
   elif st=`non critical finite singular value` then
      userinfo(2,luo,`non critical ==> discrete limit distribution`);
      userinfo(2,luo,`with constant mean and variance [`,
	`luo/cite`(`Soria90`),`, page 124]`);
      if C=Q and evalr(Signum(w0))=1 and evalr(Signum(1-w0))=1 then
      	 print(`Q(u*w) ==> derivative of geometric distribution with mean`,
	    1/(1-w0));
      elif C=`exp` and evalr(Signum(w0))=1 then		# tabular page 135
      	 print(`Poisson distribution with mean`,w0+1);
      fi;
   elif st=`non critical convergence radius non minimal` then
      userinfo(2,luo,`non critical ==> discrete limit distribution`);
      userinfo(2,luo,`with constant mean and variance [Soria90, page 138]`);
      if C=Q then 
       print(`geometric distribution with mean`,normal(w0/(1-w0)))
       elif C=L then
        print(`logarithmic distribution with mean`,normal(w0/(1-w0)/log(1/(1-w0))))
       elif C=`exp` then poissonnien(w0)
       else ERROR(`unknown constructor`,C)
      fi;
   fi
end:

is_logarithmic := proc(w,rho) # cf definition page 143
# returns true if w(z) = a*log(1/(1-z/rho)) + K + o(1/log(1/(1-z/rho)))
# when z -> rho
local d,s,mult,d1,d2,K,d3;
 if has(w,{L,ln}) then
   d := gdev(w,z=rho,2);
   if type(d,`+`) then d1:=op(1,d) else d1:=d fi;
   d1:=expand(d1);
   if match(d1=A*ln(1/(1-z/rho)^B),z,'s') then
      mult := subs(s,A*B);
      K:=glimit(w-subs(s,A*ln(1/(1-z/rho)^B)),z=rho);
      if has(K,infinity) then RETURN([false]) fi;
      d2:=w-subs(s,A*ln(1/(1-z/rho)^B))-K;
      d3:=glimit(d2 * log(1/(1-z/rho)),z=rho);
      if d3=0 then RETURN([true,mult]) fi;
   fi;
 fi;
   [false];
end:

 is_polynomial:=proc(w)
 if  type(w,polynom) then RETURN([true,degree(w,z)]) fi;
   [false];
 end:

gaussian := proc(t)
    userinfo(2,luo,Pr(alpha <= ((U-mu_n)/sigma_n <= beta)),` ---> `,
	    1/'sqrt(2*Pi)',Int(exp(-x^2/2),x=alpha..beta),
	    print(plot(exp(-x^2/2)/sqrt(2*Pi),``=-5..5)));
    `normal limit distribution with mean and variance in `.t
end:

gauss_abs := proc()
 userinfo(2,luo,print(sigma_n * Pr(U=[mu_n+lambda*sigma_n]),` ---> `,
      2/`sqrt(2*Pi)`*exp(-lambda^2/2),` for |lambda| <= lambda0`),
	print(plot(exp(-x^2/2)*2/sqrt(2*Pi),``=0..5))); 
 `absolute value of normal distribution with mean and stdev in sqrt(n)`
end:    

`print/abs` := proc(x) `|`.x.`|` end:

rayleigh := proc()
   userinfo(2,luo,
    print(sigma_n * Pr(U=[mu_n+lambda*sigma_n]),` ---> `,
      abs(lambda)*exp(-lambda^2/2),` for |lambda| <= lambda0`),
    print(plot(x*exp(-x^2/2),``=0..5)));
   `Rayleigh limit distribution with mean and standard deviation in sqrt(n)`
end:

poissonnien := proc(t)
    userinfo(2,luo,
#    print(Pr(U =k ),` ---> `,exp(- t)*t^k/k!);
	    print(plot('exp(-evalf(t))*evalf(t)^k/k!', k=0..5)));
    `poisson distribution with mean and variance in `.t
end:

is_algebraic := proc(f)
# returns true if f is of the form P = z Phi(P(z))
local eq,g;
   if type(f,function) then
      if op(0,f)=RootOf then
      	 eq := op(f);
      	 if type(eq,`+`) then
      	    map(proc(x) if x=_Z then 0 else x fi end,eq);
	    "/z;
	    if type(",polynom(rational,_Z)) or "=exp(_Z) then RETURN(true) fi
	 fi
      elif op(0,f)=W then RETURN(is_algebraic(op(f)))
      fi
   elif type(f,`*`) then
      true;
      for g in f while " do
      	 is_algebraic(g)
      od;
      RETURN(")
   elif type(f,polynom(rational,z)) then RETURN(true)
   fi;
   false
end:

search_marked := proc(f,m,l) local g;
   if not has(f,m) then NULL
   elif type(f,`+`) then
      map(proc(x,m) if has(x,m) then x fi end,[op(f)],m);
      if nops(")<>1 then ERROR(`Number of marks different from 1:`,nops("))
      else procname(op("),m,[op(l),`+`])
      fi;
   elif type(f,`*`) then
      if member(m,[op(f)]) and not has(f/m,m) then RETURN([subs(m=1,f),l]) fi;
      map(proc(x,m) if has(x,m) then x fi end,[op(f)],m);
      if nops(")<>1 then ERROR(`Number of marks different from 1:`,nops("))
      else procname(op("),m,[op(l),`*`])
      fi;
   elif type(f,function) then
      g:=op(f);
      if type(g,`*`) and member(m,[op(g)]) then RETURN([subs(m=1,f),l])
      else procname(g,m,[op(l),op(0,f)])
      fi;
   else
      ERROR(f)
   fi
end:

critical_status := proc(f,rho,rhof)
local g,rhog,taug,C;
   C:=op(0,f);
   userinfo(2,luo,`Analyzing status of constructor `.C.` at the root of`,f,`:`);
	 if evalr(Signum(rhof-rho))=1 then # rhof>rho
	    userinfo(2,luo,`convergence radius non minimal ==> non critical`);
	    RETURN(`non critical convergence radius non minimal`)
	 else # rhof=rho
       	    g:=op(f);
	    taug:=simplify(eval(subs(z=rho,g)));
	    if C=`exp` then
	       if glimit(g,z=rho)<infinity then
	       	  userinfo(2,luo,`finite singular value:`,taug,`==> exp non critical`);
		  RETURN(`non critical finite singular value`)
	       else
	       	  userinfo(2,luo,`minimal convergence radius, infinite singular value:`,taug);
		  userinfo(2,luo,`==> exp above critical`);
	       	  RETURN(`above critical`)
	       fi
	    elif member(C,[Q,L]) then
	       if evalr(Signum(normal(subsQL(taug)-1)))=0 then 
		  userinfo(2,luo,`rho=rhof=`,rho,`singular value=infinity:`,C(taug));
		  RETURN(`critical`)
	       else 
      	       	  print(`rho=rhof=`,rho,`singular value<infinity:`,C(taug));
		  RETURN(`non critical finite singular value`)
	       fi
	    else ERROR(`unknown constructor: `.C)
	    fi;
	 fi
end:

Exp := proc(x)
   if type(x,function) and op(0,x)=L then Q(op(x))
   else exp(x)
   fi;
end:

save `distribution.m`;
quit

